import numpy as np
from pathlib import Path
from importlib import resources
import pytest

from numpy.testing import (
    assert_almost_equal,
    assert_equal,
    assert_allclose,
)
from scipy.optimize._constraints import PreparedConstraint
import refnx
import refnx.analysis

# the analysis module contains the curvefitting engine
from refnx.analysis import CurveFitter, Objective

from refnx.reflect import (
    SLD,
    ReflectModel,
    Structure,
    Spline,
    Slab,
    LipidLeaflet,
    LipidLeafletGuest,
)
from refnx.reflect.structure import _profile_slicer, overall_sld
from refnx.analysis import Parameter, Interval

# the ReflectDataset object will contain the data
from refnx.dataset import ReflectDataset


class TestLipidLeaflet:
    def setup_method(self):
        self.b_h = 6.01e-4
        self.V_h = 319.0
        self.b_t = -2.92e-4
        self.V_t = 782.0
        self.APM = 60.0
        self.thick_h = 9.0
        self.thick_t = 14.0
        self.leaflet = LipidLeaflet(
            self.APM,
            self.b_h,
            self.V_h,
            self.thick_h,
            self.b_t,
            self.V_t,
            self.thick_t,
            2,
            3,
        )

        self.rho_h = self.b_h / self.V_h / 1e-6
        self.rho_t = self.b_t / self.V_t / 1e-6
        self.phi_solv_h = 1 - (self.V_h / (self.APM * self.thick_h))
        self.phi_solv_t = 1 - (self.V_t / (self.APM * self.thick_t))

    def test_slabs(self):
        # check that slab calculation from parameters is correct
        slabs = self.leaflet.slabs()
        theoretical = np.array(
            [
                [self.thick_h, self.rho_h, 0, 3, self.phi_solv_h],
                [self.thick_t, self.rho_t, 0, 2, self.phi_solv_t],
            ]
        )
        assert_allclose(slabs, theoretical, rtol=1e-15)

        # check that we can flip the lipid leaflet
        self.leaflet.reverse_monolayer = True
        theoretical = np.flipud(theoretical)
        theoretical[:, 3] = theoretical[::-1, 3]
        assert_allclose(self.leaflet.slabs(), theoretical, rtol=1e-15)

    def test_solvent_penetration(self):
        # check different types of solvation for heads/tails.
        self.leaflet.head_solvent = SLD(1.23)
        slabs = self.leaflet.slabs()
        assert_allclose(
            slabs[0, 1],
            1.23 * self.phi_solv_h + (1 - self.phi_solv_h) * self.rho_h,
        )
        assert_allclose(slabs[0, 4], 0)

        self.leaflet.head_solvent = None
        self.leaflet.tail_solvent = SLD(1.23)
        slabs = self.leaflet.slabs()
        assert_allclose(
            slabs[1, 1],
            1.23 * self.phi_solv_t + (1 - self.phi_solv_t) * self.rho_t,
        )
        assert_allclose(slabs[1, 4], 0)

    def test_initialisation_with_SLD(self):
        # we should be able to initialise with SLD objects
        heads = SLD(6.01e-4 + 0j)
        tails = SLD(-2.92e-4 + 0j)
        new_leaflet = LipidLeaflet(
            self.APM,
            heads,
            self.V_h,
            self.thick_h,
            tails,
            self.V_t,
            self.thick_t,
            2,
            3,
        )
        slabs = self.leaflet.slabs()
        new_slabs = new_leaflet.slabs()
        assert_allclose(new_slabs, slabs)

    def test_repr(self):
        # test that we can reconstruct the object from a repr
        s = repr(self.leaflet)
        q = eval(s)
        assert_equal(q.slabs(), self.leaflet.slabs())

    def test_lipidleafletguest_no_solvent_specified(self):
        phi_guest_t = Parameter(0.1)
        sld_guest = SLD(7.6)
        sld_solvent = SLD(5.55)
        with pytest.warns(RuntimeWarning):
            leaflet = LipidLeafletGuest(
                self.APM,
                self.b_h,
                self.V_h,
                self.thick_h,
                self.b_t,
                self.V_t,
                self.thick_t,
                2,
                3,
                0,
                phi_guest_t,
                sld_guest,
            )
        assert isinstance(leaflet, LipidLeafletGuest)
        assert leaflet.phi_guest_t is phi_guest_t
        assert leaflet.sld_guest is sld_guest

        # check volume fractions are calculated correctly
        vft = self.V_t / (self.APM * self.thick_t)
        assert_allclose(leaflet.volfrac_t, vft)
        vfh = self.V_h / (self.APM * self.thick_h)
        assert_allclose(leaflet.volfrac_h, vfh)

        vftg = (1 - vft) * phi_guest_t.value
        assert_equal(leaflet.volfrac_guest_t, vftg)

        # check slab representation
        slabs = leaflet.slabs()
        assert_allclose(slabs[0, 0], self.thick_h)
        assert_allclose(slabs[1, 0], self.thick_t)
        assert_allclose(slabs[0, -1], 1 - vfh)
        vfsolv = 1 - vft - vftg
        assert_allclose(slabs[1, -1], vfsolv)
        sld_h = self.b_h / self.V_h * 1e6
        assert_allclose(slabs[0, 1], sld_h)

        sld_t = self.b_t / self.V_t * 1e6
        _solvated_slabs = overall_sld(np.copy(slabs), complex(sld_solvent))
        assert_allclose(
            _solvated_slabs[1, 1],
            vft * sld_t
            + vftg * complex(sld_guest).real
            + vfsolv * complex(sld_solvent).real,
        )
        den = vft + vftg
        _sld = (vft * sld_t + vftg * sld_guest.real.value) / den
        assert_allclose(slabs[1, 1], _sld)

    def test_lipidleafletguest_solvent_specified(self):
        phi_guest_t = Parameter(0.1)
        sld_guest = SLD(7.6)
        sld_solvent = SLD(5.55)

        with pytest.warns(RuntimeWarning):
            leaflet = LipidLeafletGuest(
                self.APM,
                self.b_h,
                self.V_h,
                self.thick_h,
                self.b_t,
                self.V_t,
                self.thick_t,
                2,
                3,
                0,
                phi_guest_t,
                sld_guest,
                head_solvent=sld_solvent,
                tail_solvent=sld_solvent,
            )
        # check volume fractions are calculated correctly
        vft = self.V_t / (self.APM * self.thick_t)
        vfh = self.V_h / (self.APM * self.thick_h)
        vftg = (1 - vft) * phi_guest_t.value
        assert_equal(leaflet.volfrac_guest_t, vftg)

        # check slab representation
        slabs = leaflet.slabs()
        assert_allclose(slabs[0, 0], self.thick_h)
        assert_allclose(slabs[1, 0], self.thick_t)
        # should be zero because solvation is done by the component
        # and not by parent structure
        assert_allclose(slabs[0, -1], 0)
        assert_allclose(slabs[1, -1], 0)
        sld_h = self.b_h / self.V_h * 1e6
        assert_allclose(
            slabs[0, 1], sld_h * vfh + (1 - vfh) * complex(sld_solvent).real
        )

        vfsolv = 1 - vft - vftg
        sld_t = self.b_t / self.V_t * 1e6
        _solvated_slabs = overall_sld(np.copy(slabs), complex(sld_solvent))
        _sld_tail = (
            vft * sld_t
            + vftg * complex(sld_guest).real
            + vfsolv * complex(sld_solvent).real
        )
        assert_allclose(
            _solvated_slabs[1, 1],
            _sld_tail,
        )
        assert_allclose(slabs[1, 1], _sld_tail)


def test_lipid_leaflet_example():
    pth = resources.files(refnx.analysis) / "tests"

    data_d2o = ReflectDataset(pth / "c_PLP0016596.dat")
    data_d2o.name = "d2o"

    si = SLD(2.07 + 0j)
    sio2 = SLD(3.47 + 0j)

    # the following represent the solvent contrasts used in the experiment
    d2o = SLD(6.36 + 0j)

    # We want the `real` attribute parameter to vary in the analysis, and we want to apply
    # uniform bounds. The `setp` method of a Parameter is a way of changing many aspects of
    # Parameter behaviour at once.
    d2o.real.setp(vary=True, bounds=(6.1, 6.36))
    d2o.real.name = "d2o SLD"

    # Parameter for the area per molecule each DMPC molecule occupies at the surface. We
    # use the same area per molecule for the inner and outer leaflets.
    apm = Parameter(56, "area per molecule", vary=True, bounds=(52, 65))

    # the sum of scattering lengths for the lipid head and tail in Angstrom.
    b_heads = Parameter(6.01e-4, "b_heads")
    b_tails = Parameter(-2.92e-4, "b_tails")

    # the volume occupied by the head and tail groups in cubic Angstrom.
    v_heads = Parameter(319, "v_heads")
    v_tails = Parameter(782, "v_tails")

    # the head and tail group thicknesses.
    inner_head_thickness = Parameter(
        9, "inner_head_thickness", vary=True, bounds=(4, 11)
    )
    outer_head_thickness = Parameter(
        9, "outer_head_thickness", vary=True, bounds=(4, 11)
    )
    tail_thickness = Parameter(
        14, "tail_thickness", vary=True, bounds=(10, 17)
    )

    # finally construct a `LipidLeaflet` object for the inner and outer leaflets.
    # Note that here the inner and outer leaflets use the same area per molecule,
    # same tail thickness, etc, but this is not necessary if the inner and outer
    # leaflets are different.
    inner_leaflet = LipidLeaflet(
        apm,
        b_heads,
        v_heads,
        inner_head_thickness,
        b_tails,
        v_tails,
        tail_thickness,
        3,
        3,
    )

    # we reverse the monolayer for the outer leaflet because the tail groups face upwards
    outer_leaflet = LipidLeaflet(
        apm,
        b_heads,
        v_heads,
        outer_head_thickness,
        b_tails,
        v_tails,
        tail_thickness,
        3,
        0,
        reverse_monolayer=True,
    )

    # Slab constructed from SLD object.
    sio2_slab = sio2(15, 3)
    sio2_slab.thick.setp(vary=True, bounds=(2, 30))
    sio2_slab.thick.name = "sio2 thickness"
    sio2_slab.rough.setp(vary=True, bounds=(0, 7))
    sio2_slab.rough.name = "sio2 roughness"
    sio2_slab.vfsolv.setp(0.1, vary=True, bounds=(0.0, 0.5))
    sio2_slab.vfsolv.name = "sio2 solvation"

    solv_roughness = Parameter(3, "bilayer/solvent roughness")
    solv_roughness.setp(vary=True, bounds=(0, 5))

    s_d2o = (
        si | sio2_slab | inner_leaflet | outer_leaflet | d2o(0, solv_roughness)
    )

    model_d2o = ReflectModel(s_d2o)

    model_d2o.scale.setp(vary=True, bounds=(0.9, 1.1))
    model_d2o.bkg.setp(vary=True, bounds=(-1e-6, 1e-6))
    objective_d2o = Objective(model_d2o, data_d2o)

    con_inner = inner_leaflet.make_constraint(objective_d2o)
    con_outer = outer_leaflet.make_constraint(objective_d2o)

    fitter = CurveFitter(objective_d2o)

    fitter.fit(
        "differential_evolution",
        constraints=(con_inner, con_outer),
        polish=False,
        popsize=10,
    )
    assert inner_leaflet.volfrac_h <= 1
    assert inner_leaflet.volfrac_t <= 1

    assert outer_leaflet.volfrac_h <= 1
    assert outer_leaflet.volfrac_t <= 1

    arr = np.array(objective_d2o.parameters)
    pc = PreparedConstraint(con_inner, arr)
    v1 = pc.violation(arr)
    apm.value = 20.0
    arr = np.array(objective_d2o.parameters)
    v2 = pc.violation(arr)
    assert (v2 > 0).all()

    assert not np.allclose(v1, v2)



import numpy as np
from refnx.analysis import (
    possibly_create_parameter,
    Parameter,
    is_parameter,
    Parameters,
)
from refnx.reflect import Component, Structure, SLD


class FunctionalForm(Component):
    """
    Component used to describe an analytic SLD profile.

    Parameters
    ----------
    extent : Parameter or float
        The total extent of the functional region
    profile : callable
        callable returning the SLD of the functional form. Has the signature
        ``profile(z, extent, left_sld, right_sld, **kwds)`` where ``z`` is an
        array specifying the distances at which the functional form should be
        calculated, ``extent`` is a float specifying the total extent of the
        Component, ``left_sld`` and ``right_sld`` are the complex SLDs of the
        Components preceding and following this Component. ``kwds`` are extra
        parameters used to calculate the shape of the functional form.
        The function should return a tuple, ``(slds, vfsolv)`` where ``slds``
        is an array (possibly of type np.complex128) of shape `(len(z),)`.
        Similarly, ``vfsolv`` is an array of shape `(len(z))` specifying the
        volume fraction of solvent at each distance in ``z``. If ``vfsolv`` is
        not necessary then return ``(slds, None)``.
    name : str
        Name of component
    microslab_max_thickness : float, optional
        Thickness of microslicing of spline for reflectivity calculation
    kwds : dict
        Named keywords that supply extra Parameters to the ``profile`` callable.
        These parameters are passed numerically, not as Parameter objects.

    Examples
    --------
    A linear ramp. Note that the `dummy_param` is not actually used anywhere.

    >>> def line(z, extent, left_sld, right_sld, dummy_param=None):
    ...     grad = (right_sld - left_sld) / extent
    ...     intercept = left_sld
    ...     # we don't calculate the volume fraction of solvent
    ...     return z*grad*dummy_param + intercept, None

    >>> si = SLD(2.07)
    >>> d2o = SLD(6.36)
    >>> p = Parameter(1)

    >>> form = FunctionalForm(100, line, dummy_param=p)
    >>> s = si | form | d2o(0, 3)

    A quadratic example that goes through the two end points

    >>> def quadratic(z, extent, left_sld, right_sld, x=None, y=None):
    ...     res = np.polyfit(
    ...         [0., x, extent],
    ...         [np.real(left_sld), y, np.real(right_sld)],
    ...         deg=2
    ...     )
    ...     return np.polyval(res, z), None

    >>> si = SLD(2.07)
    >>> d2o = SLD(6.36)
    >>> x = Parameter(4.)
    >>> y = Parameter(5.)
    >>> quad = FunctionalForm(100., quadratic, x=x, y=y)
    >>> s = si | quad | d2o(0, 3)

    """

    def __init__(
        self, extent, profile, name=None, microslab_max_thickness=1, **kwds
    ):
        super(FunctionalForm, self).__init__(name=name)
        self.profile = profile
        self.other_params = {k: v for k, v in kwds.items() if is_parameter(v)}

        self.extent = possibly_create_parameter(extent)
        self.microslab_max_thickness = microslab_max_thickness

    @property
    def parameters(self):
        p = Parameters(name=self.name)
        p.extend([self.extent])
        p.extend(list(self.other_params.values()))
        return p

    def slabs(self, structure=None):
        assert (
            structure is not None
        ), "In order to calculate slab representation please provide a Structure"

        try:
            loc = structure.index(self)
            # figure out SLDs for the bracketing Components.
            left_component = structure[loc - 1]
            right_component = structure[(loc + 1) % len(structure)]
        except ValueError:
            raise ValueError(
                "FunctionalForm didn't appear to be part of a super structure"
            )

        left_slab = structure.overall_sld(
            np.atleast_2d(left_component.slabs(structure)[-1]),
            structure.solvent,
        )
        left_sld = complex(left_slab[..., 1][0], left_slab[..., 2][0])

        right_slab = structure.overall_sld(
            np.atleast_2d(right_component.slabs(structure)[0]),
            structure.solvent,
        )
        right_sld = complex(right_slab[..., 1][0], right_slab[..., 2][0])

        num_slabs = int(
            np.ceil(float(self.extent) / self.microslab_max_thickness)
        )
        slab_thick = self.extent.value / num_slabs

        slabs = np.zeros((num_slabs, 5))
        slabs[..., 0] = slab_thick

        dist = np.cumsum(slabs[..., 0]) - 0.5 * slab_thick

        pars = {k: float(v) for k, v in self.other_params.items()}
        res, vfsolv = self.profile(
            dist, self.extent.value, left_sld, right_sld, **pars
        )

        slabs[..., 1] = np.real(res)
        slabs[..., 2] = np.imag(res)

        if vfsolv is not None:
            slabs[..., 4] = np.clip(vfsolv, 0, 1)

        return slabs



# contents of conftest.py
import zipfile
import shutil
import urllib.request
from pathlib import Path

import pytest

import refnx


@pytest.fixture(scope="session")
def data_directory(tmp_path_factory):
    """
    Retrieves the refnx-testdata repository, placing it in a temporary
    directory, for use in pytest fixtures

    Returns
    -------
    data_dir: str or None
        If the retrieval works then a str pointing to the test data is
        returned. If the retrieval fails then None is returned.
    """
    url = "https://github.com/refnx/refnx-testdata/archive/master.zip"
    tmpdir = tmp_path_factory.mktemp("data")

    try:
        # grab the test data
        with (
            urllib.request.urlopen(url, timeout=5) as response,
            open(tmpdir / "master.zip", "wb") as f,
        ):
            shutil.copyfileobj(response, f)

        # master.zip is in tmpdir
        with zipfile.ZipFile(tmpdir / "master.zip") as zf:
            zf.extractall(path=tmpdir)

        data_dir = tmpdir / "refnx-testdata-master" / "data"
    except (urllib.error.URLError, TimeoutError):
        pytest.skip("No data directory available")

    return data_dir



# -*- coding: utf-8 -*-

from __future__ import (
    division,
    print_function,
    absolute_import,
    unicode_literals,
)

__all__ = ["make_ladder", "Sampler"]

import operator
import attr
import itertools
import numpy as np

from . import util, chain, ensemble


def make_ladder(ndim, ntemps=None, Tmax=None):
    """
    Returns a ladder of :math:`\\beta \\equiv 1/T` under a geometric spacing that is determined by the
    arguments ``ntemps`` and ``Tmax``.  The temperature selection algorithm works as follows:

    Ideally, ``Tmax`` should be specified such that the tempered posterior looks like the prior at
    this temperature.  If using adaptive parallel tempering, per `arXiv:1501.05823
    <http://arxiv.org/abs/1501.05823>`_, choosing ``Tmax = inf`` is a safe bet, so long as
    ``ntemps`` is also specified.

    :param ndim:
        The number of dimensions in the parameter space.

    :param ntemps: (optional)
        If set, the number of temperatures to generate.

    :param Tmax: (optional)
        If set, the maximum temperature for the ladder.

    Temperatures are chosen according to the following algorithm:

    * If neither ``ntemps`` nor ``Tmax`` is specified, raise an exception (insufficient
      information).
    * If ``ntemps`` is specified but not ``Tmax``, return a ladder spaced so that a Gaussian
      posterior would have a 25% temperature swap acceptance ratio.
    * If ``Tmax`` is specified but not ``ntemps``:

      * If ``Tmax = inf``, raise an exception (insufficient information).
      * Else, space chains geometrically as above (for 25% acceptance) until ``Tmax`` is reached.

    * If ``Tmax`` and ``ntemps`` are specified:

      * If ``Tmax = inf``, place one chain at ``inf`` and ``ntemps-1`` in a 25% geometric spacing.
      * Else, use the unique geometric spacing defined by ``ntemps`` and ``Tmax``.

    """

    if type(ndim) != int or ndim < 1:
        raise ValueError("Invalid number of dimensions specified.")
    if ntemps is None and Tmax is None:
        raise ValueError("Must specify one of ``ntemps`` and ``Tmax``.")
    if Tmax is not None and Tmax <= 1:
        raise ValueError("``Tmax`` must be greater than 1.")
    if ntemps is not None and (type(ntemps) != int or ntemps < 1):
        raise ValueError("Invalid number of temperatures specified.")

    tstep = np.array(
        [
            25.2741,
            7.0,
            4.47502,
            3.5236,
            3.0232,
            2.71225,
            2.49879,
            2.34226,
            2.22198,
            2.12628,
            2.04807,
            1.98276,
            1.92728,
            1.87946,
            1.83774,
            1.80096,
            1.76826,
            1.73895,
            1.7125,
            1.68849,
            1.66657,
            1.64647,
            1.62795,
            1.61083,
            1.59494,
            1.58014,
            1.56632,
            1.55338,
            1.54123,
            1.5298,
            1.51901,
            1.50881,
            1.49916,
            1.49,
            1.4813,
            1.47302,
            1.46512,
            1.45759,
            1.45039,
            1.4435,
            1.4369,
            1.43056,
            1.42448,
            1.41864,
            1.41302,
            1.40761,
            1.40239,
            1.39736,
            1.3925,
            1.38781,
            1.38327,
            1.37888,
            1.37463,
            1.37051,
            1.36652,
            1.36265,
            1.35889,
            1.35524,
            1.3517,
            1.34825,
            1.3449,
            1.34164,
            1.33847,
            1.33538,
            1.33236,
            1.32943,
            1.32656,
            1.32377,
            1.32104,
            1.31838,
            1.31578,
            1.31325,
            1.31076,
            1.30834,
            1.30596,
            1.30364,
            1.30137,
            1.29915,
            1.29697,
            1.29484,
            1.29275,
            1.29071,
            1.2887,
            1.28673,
            1.2848,
            1.28291,
            1.28106,
            1.27923,
            1.27745,
            1.27569,
            1.27397,
            1.27227,
            1.27061,
            1.26898,
            1.26737,
            1.26579,
            1.26424,
            1.26271,
            1.26121,
            1.25973,
        ]
    )

    if ndim > tstep.shape[0]:
        # An approximation to the temperature step at large
        # dimension
        tstep = 1.0 + 2.0 * np.sqrt(np.log(4.0)) / np.sqrt(ndim)
    else:
        tstep = tstep[ndim - 1]

    appendInf = False
    if Tmax == np.inf:
        appendInf = True
        Tmax = None
        ntemps = ntemps - 1

    if ntemps is not None:
        if Tmax is None:
            # Determine Tmax from ntemps.
            Tmax = tstep ** (ntemps - 1)
    else:
        if Tmax is None:
            raise ValueError(
                "Must specify at least one of ``ntemps"
                " and "
                "finite ``Tmax``."
            )

        # Determine ntemps from Tmax.
        ntemps = int(np.log(Tmax) / np.log(tstep) + 2)

    betas = np.logspace(0, -np.log10(Tmax), ntemps)
    if appendInf:
        # Use a geometric spacing, but replace the top-most temperature with
        # infinity.
        betas = np.concatenate((betas, [0]))

    return betas


@attr.s(slots=True, frozen=True)
class LikePriorEvaluator(object):
    """
    Wrapper class for logl and logp.

    """

    logl = attr.ib()
    logp = attr.ib()
    logl_args = attr.ib(factory=list)
    logp_args = attr.ib(factory=list)
    logl_kwargs = attr.ib(factory=dict)
    logp_kwargs = attr.ib(factory=dict)

    def __call__(self, x):
        lp = self.logp(x, *self.logp_args, **self.logp_kwargs)
        if np.isnan(lp):
            raise ValueError("Prior function returned NaN.")

        if lp == float("-inf"):
            # Can't return -inf, since this messes with beta=0 behaviour.
            ll = 0
        else:
            ll = self.logl(x, *self.logl_args, **self.logl_kwargs)
            if np.isnan(ll).any():
                raise ValueError("Log likelihood function returned NaN.")

        return ll, lp


@attr.s(slots=True, frozen=True)
class Sampler(object):
    # Mandatory parameters.
    nwalkers = attr.ib(converter=int)
    ndim = attr.ib(converter=int)

    logl = attr.ib()
    logp = attr.ib()
    logl_args = attr.ib(converter=list, factory=list)
    logp_args = attr.ib(converter=list, factory=list)
    logl_kwargs = attr.ib(converter=dict, factory=dict)
    logp_kwargs = attr.ib(converter=dict, factory=dict)

    betas = attr.ib(default=None)

    # Tuning parameters.
    adaptive = attr.ib(converter=bool, default=False)
    adaptation_lag = attr.ib(converter=int, default=10000)
    adaptation_time = attr.ib(converter=int, default=100)
    scale_factor = attr.ib(converter=float, default=2)

    _mapper = attr.ib(default=map)
    _evaluator = attr.ib(type=LikePriorEvaluator, init=False, default=None)
    _data = attr.ib(type=np.ndarray, init=False, default=None)

    @nwalkers.validator
    def _validate_nwalkers(self, attribute, value):
        if value % 2 != 0:
            raise ValueError("The number of walkers must be even.")
        if value < 2 * self.ndim:
            raise ValueError(
                "The number of walkers must be greater than 2 * dimension."
            )

    @ndim.validator
    def _validate_ndim(self, attribute, value):
        if value < 1:
            raise ValueError("Number of dimensions must be positive.")

    @betas.validator
    def _validate_betas(self, attribute, value):
        try:
            # see if betas is an integer.
            v = operator.index(value)
            if v < 1:
                raise ValueError("Need at least one temperature")
            return
        except TypeError:
            pass

        if len(value) < 1:
            raise ValueError("Need at least one temperature!")
        if (value < 0).any():
            raise ValueError("Temperatures must be non-negative.")

    @logl.validator
    @logp.validator
    def _is_callable(self, attribute, value):
        if not callable(value):
            raise TypeError("{} must be callable".format(attribute.name))

    def __attrs_post_init__(self):
        if self.betas is None:
            betas = make_ladder(self.ndim)
        elif isinstance(self.betas, int):
            # Treat this as the number of temperatures to use.
            betas = make_ladder(self.ndim, self.betas)
        else:
            betas = util._ladder(self.betas)

        object.__setattr__(self, "betas", betas)
        object.__setattr__(
            self,
            "_evaluator",
            LikePriorEvaluator(
                logl=self.logl,
                logp=self.logp,
                logl_args=self.logl_args,
                logp_args=self.logp_args,
                logl_kwargs=self.logl_kwargs,
                logp_kwargs=self.logp_kwargs,
            ),
        )

    def ensemble(self, x, rng=None):
        random = np.random.default_rng(rng)

        config = ensemble.EnsembleConfiguration(
            adaptation_lag=self.adaptation_lag,
            adaptation_time=self.adaptation_time,
            scale_factor=self.scale_factor,
            evaluator=self._evaluator,
        )
        return ensemble.Ensemble(
            x=x,
            betas=self.betas.copy(),
            config=config,
            adaptive=self.adaptive,
            rng=random,
            mapper=self._mapper,
        )

    def sample(self, x, rng=None, thin_by=None):
        """
        Return a stateless iterator.

        """

        if thin_by is None:
            thin_by = 1

        # Don't yield the starting state.
        ensemble = self.ensemble(x, rng)
        while True:
            for _ in range(thin_by):
                ensemble.step()
            yield ensemble

    def chain(self, x, rng=None, thin_by=None):
        """
        Create a stateful chain that stores its history.

        """
        return chain.Chain(self.ensemble(x, rng), thin_by)



"""
*Calculates the specular (Neutron or X-ray) reflectivity from a stratified
series of layers.

The refnx code is distributed under the following license:

Copyright (c) 2015 A. R. J. Nelson, ANSTO

Permission to use and redistribute the source code or binary forms of this
software and its documentation, with or without modification is hereby
granted provided that the above notice of copyright, these terms of use,
and the disclaimer of warranty below appear in the source code and
documentation, and that none of the names of above institutions or
authors appear in advertising or endorsement of works derived from this
software without specific prior written permission from all parties.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THIS SOFTWARE.

"""

from functools import reduce

import numpy as np
import jax.numpy as jnp
from jax import jit
from refnx.reflect.reflect_model import gauss_legendre

TINY = 1e-30

_FWHM = 2 * np.sqrt(2 * np.log(2.0))
_INTLIMIT = 3.5


def jabeles(q, layers, scale=1.0, bkg=0, threads=0):
    # Note, this code must have 64 bit operation enabled before it is used
    # You can do this with:
    # from jax import config
    # config.update("jax_enable_x64", True)

    qvals = q.astype(jnp.float64)
    flatq = qvals.ravel()

    nlayers = layers.shape[0] - 2
    npnts = flatq.size

    mi00 = jnp.ones((npnts, nlayers + 1), jnp.complex128)

    sld = jnp.zeros(nlayers + 2, jnp.complex128)

    # addition of TINY is to ensure the correct branch cut
    # in the complex sqrt calculation of kn.
    sld = sld.at[1:].add(
        ((layers[1:, 1] - layers[0, 1]) + 1j * (jnp.abs(layers[1:, 2]) + TINY))
        * 1.0e-6
    )

    kn = jnp.sqrt(flatq[:, jnp.newaxis] ** 2.0 / 4.0 - 4.0 * jnp.pi * sld)
    # reflectances for each layer
    # rj.shape = (npnts, nlayers + 1)
    damping = jnp.exp(-2.0 * kn[:, :-1] * kn[:, 1:] * layers[1:, 3] ** 2)
    rj = (kn[:, :-1] - kn[:, 1:]) / (kn[:, :-1] + kn[:, 1:]) * damping

    # characteristic matrices for each layer
    # miNN.shape = (npnts, nlayers + 1)
    if nlayers:
        mi00 = mi00.at[:, 1:].set(
            jnp.exp(kn[:, 1:-1] * 1j * jnp.fabs(layers[1:-1, 0]))
        )
    mi11 = 1.0 / mi00
    mi10 = rj * mi00
    mi01 = rj * mi11

    # initialise matrix total
    mrtot00 = mi00[:, 0]
    mrtot01 = mi01[:, 0]
    mrtot10 = mi10[:, 0]
    mrtot11 = mi11[:, 0]

    for _mi00, _mi10, _mi01, _mi11 in zip(
        mi00[:, 1:].T, mi10[:, 1:].T, mi01[:, 1:].T, mi11[:, 1:].T
    ):
        # matrix multiply mrtot by characteristic matrix
        p00 = mrtot00 * _mi00 + mrtot10 * _mi01
        p10 = mrtot00 * _mi10 + mrtot10 * _mi11
        p01 = mrtot01 * _mi00 + mrtot11 * _mi01
        p11 = mrtot01 * _mi10 + mrtot11 * _mi11

        mrtot00, mrtot01, mrtot10, mrtot11 = p00, p01, p10, p11

    r = mrtot01 / mrtot00
    reflectivity = r * jnp.conj(r)

    return scale * jnp.real(jnp.reshape(reflectivity, qvals.shape)) + bkg


# abeles_jax = jabeles
abeles_jax = jit(jabeles)


def jax_smeared_kernel_pointwise(qvals, w, dqvals, quad_order=17, threads=0):
    # get the gauss-legendre weights and abscissae
    abscissa, weights = gauss_legendre(quad_order)

    # get the normal distribution at that point
    prefactor = 1.0 / np.sqrt(2 * np.pi)

    def gauss(x):
        return np.exp(-0.5 * x * x)

    gaussvals = prefactor * gauss(abscissa * _INTLIMIT)

    # integration between -3.5 and 3.5 sigma
    va = qvals - _INTLIMIT * dqvals / _FWHM
    vb = qvals + _INTLIMIT * dqvals / _FWHM

    va = va[:, np.newaxis]
    vb = vb[:, np.newaxis]

    qvals_for_res = (np.atleast_2d(abscissa) * (vb - va) + vb + va) / 2.0
    smeared_rvals = abeles_jax(qvals_for_res, w)

    smeared_rvals = np.reshape(smeared_rvals, (qvals.size, abscissa.size))

    smeared_rvals *= np.atleast_2d(gaussvals * weights)
    return np.sum(smeared_rvals, 1) * _INTLIMIT



import operator
from types import MethodType
from collections import UserList

import numpy as np
from refnx._lib import flatten, unique as f_unique
from refnx.analysis import Interval, PDF, Bounds


# Functions for making Functors
def MAKE_BINARY(opfn):
    return lambda self, other: (_BinaryOp(self, asMagicNumber(other), opfn))


def MAKE_RBINARY(opfn):
    return lambda self, other: (_BinaryOp(asMagicNumber(other), self, opfn))


def MAKE_UNARY(opfn):
    return lambda val: _UnaryOp(val, opfn)


def asMagicNumber(x):
    return x if isinstance(x, BaseParameter) else Constant(x)


# a function that takes a function that returns a function
# MAKE_BINARY = lambda opfn: lambda self, other: (
#     _BinaryOp(self, asMagicNumber(other), opfn))
# MAKE_RBINARY = lambda opfn: lambda self, other: (
#     _BinaryOp(asMagicNumber(other), self, opfn))
# MAKE_UNARY = lambda opfn: lambda val: _UnaryOp(val, opfn)
# asMagicNumber = lambda x: x if isinstance(x, BaseParameter) else Constant(x)

# mathematical operations that we might want to use in constraints
ops = {
    "sin": np.sin,
    "cos": np.cos,
    "tan": np.tan,
    "arcsin": np.arcsin,
    "arccos": np.arccos,
    "arctan": np.arctan,
    "log": np.log,
    "log10": np.log10,
    "exp": np.exp,
    "sqrt": np.sqrt,
    "sum": np.sum,
}
math_ops = {k: MAKE_UNARY(v) for k, v in ops.items()}

binary = [
    operator.add,
    operator.sub,
    operator.mul,
    operator.truediv,
    operator.floordiv,
    np.power,
    operator.pow,
    operator.mod,
]

unary = [
    operator.neg,
    operator.abs,
    np.sin,
    np.tan,
    np.cos,
    np.arcsin,
    np.arctan,
    np.arccos,
    np.log10,
    np.log,
    np.sqrt,
    np.exp,
]


class Parameters(UserList):
    """
    A sequence of Parameters

    Parameters
    ----------
    data : sequence
        A sequence of :class:`Parameter` or :class:`Parameters`
    name : str
        Name of this :class:`Parameters` instance
    """

    def __init__(self, data=(), name=None):
        super().__init__()
        self.name = name
        self.data.extend(data)

    def __getitem__(self, i):
        if type(i) is str:
            try:
                _names = [param.name for param in self.data]
                idx = _names.index(i)
            except ValueError:
                return None
            finally:
                return self.data[idx]
        else:
            return self.data[i]

    def __setitem__(self, i, v):
        # i can be index or string
        # v has to be Parameters or Parameter
        if type(i) is str:
            _names = [param.name for param in self.data]
            # this will raise a KeyError if the key isn't in there.
            idx = _names.index(i)
            self.data[idx] = v
        else:
            self.data[i] = v

    def __repr__(self):
        return f"Parameters(data={self.data!r}, name={self.name!r})"

    def __str__(self):
        s = list()
        s.append(f"{'':_>80}")
        s.append(f"Parameters: {self.name!r: ^15}")

        for el in self._pprint():
            s.append(el)

        return "\n".join(list(flatten(s)))

    def _pprint(self):
        for el in self.data:
            if is_parameters(el):
                yield str(el)
            else:
                yield str(el)

    def __contains__(self, item):
        """
        Does this instance contain a given :class:`Parameter`
        """
        return id(item) in [id(p) for p in f_unique(flatten(self.data))]

    def __ior__(self, other):
        """
        Concatenate Parameter(s). You can concatenate :class:`Parameters` with
        :class:`Parameter` or :class:`Parameters` instances.
        """
        # self |= other
        if not (is_parameter(other) or is_parameters(other)):
            raise ValueError(
                "Can only concatenate a Parameter with another"
                " Parameter or Parameters instance"
            )
        self.append(other)
        return self

    def __or__(self, other):
        """
        concatenate Parameter(s). You can concatenate :class:`Parameters` with
        :class:`Parameter` or :class:`Parameters` instances.
        """
        # c = self | other
        if not (is_parameter(other) or is_parameters(other)):
            raise ValueError(
                "Can only concatenate a Parameter with another"
                " Parameter or Parameters instance"
            )
        c = Parameters(self.data, self.name)
        c.append(other)
        return c

    def logp(self):
        """
        Calculates logp for all the parameters

        Returns
        -------
        logp : float
            Log probability for all the parameters
        """
        # logp for all the parameters
        return np.sum(
            [
                param.logp()
                for param in f_unique(flatten(self.data))
                if param.vary
            ]
        )

    def __array__(self, dtype=None, copy=None):
        """
        Convert Parameters to an array containing their numerical values.

        The dtype and copy keywords are ignored.
        """
        return np.array([float(p) for p in flatten(self.data)])

    @property
    def pvals(self):
        """
        An array containing the values of all the :class:`Parameter` in this
        object.
        """
        return np.array(self)

    @pvals.setter
    def pvals(self, pvals):
        varying = [
            param for param in f_unique(flatten(self.data)) if param.vary
        ]
        if np.size(pvals) == len(varying):
            [
                setattr(param, "value", pvals[i])
                for i, param in enumerate(varying)
            ]
            return

        flattened_parameters = list(flatten(self.data))

        if np.size(pvals) == len(flattened_parameters):
            [
                setattr(param, "value", pvals[i])
                for i, param in enumerate(flattened_parameters)
            ]
            return
        else:
            raise ValueError(
                "You supplied the wrong number of values %d when "
                "setting this Parameters.pvals attribute" % len(pvals)
            )

    @property
    def parameters(self):
        return self

    @property
    def nparam(self):
        return len(self.flattened())

    def flattened(self, unique=False):
        """
        A list of all the :class:`Parameter` contained in this object,
        including those contained within :class:`Parameters` at any depth.

        Parameters
        ----------
        unique : bool
            The list will only contain unique objects.

        Returns
        -------
        params : list
            A list of :class:`Parameter` contained in this object.

        """
        if unique:
            return list(f_unique(flatten(self.data)))
        else:
            return list(flatten(self.data))

    def names(self):
        """
        Returns
        -------
        names : list
            A list of all the names of all the :class:`Parameter` contained in
            this object.
        """
        return [param.name for param in flatten(self.data)]

    def nvary(self):
        """
        Returns
        -------
        nvary : int
            The number of :class:`Parameter` contained in this object that are
            allowed to vary.

        """
        return len([1 for param in f_unique(flatten(self.data)) if param.vary])

    def constrained_parameters(self):
        """
        Returns
        -------
        constrained_parameters : list
            A list of unique :class:`Parameter` contained in this object that
            have constraints.
        """
        return [
            param
            for param in f_unique(flatten(self.data))
            if param.constraint is not None
        ]

    def varying_parameters(self):
        """
        Unique list of varying parameters

        Returns
        -------
        p : list
            Unique list of varying parameters
        """
        lst = []
        for p in flatten(self.data):
            if p.vary:
                lst.append(p)
                continue
            if len(p._deps):
                lst.extend([_p for _p in p.dependencies() if _p.vary])
        # should already be totally flattened by this point
        return Parameters(f_unique(lst))

    def pgen(self, ngen=1000, nburn=0, nthin=1, random_state=None):
        """
        Yield random parameter vectors from MCMC samples.

        Parameters
        ----------
        ngen : int, optional
            the number of samples to yield. The actual number of samples
            yielded is `min(ngen, chain.size)`
        nburn : int, optional
            discard this many steps from the start of the chain
        nthin : int, optional
            only accept every `nthin` samples from the chain
        random_state : {int, np.random.Generator, None}
            random number generator that picks the samples

        Yields
        ------
        pvec : np.ndarray
            A randomly chosen parameter vector

        Notes
        -----
        The entire parameter vector is yielded, not only the varying
        parameters. The reason for this is that some parameters may possess a
        chain if they are not varying, because they are controlled by a
        constraint.
        """

        # it's still possible to have chains, even if there are no varying
        # parameters, if there are parameters that have constraints
        # generate for all params that have chains.
        chain_pars = [
            i for i, p in enumerate(self.flattened()) if p.chain is not None
        ]

        chains = np.array(
            [
                np.ravel(param.chain[..., nburn::nthin])
                for param in self.flattened()
                if param.chain is not None
            ]
        )

        if len(chains) == 0 or np.size(chains, 1) == 0:
            raise ValueError("There were no chains to sample from")

        samples = np.arange(np.size(chains, 1))

        rng = np.random.default_rng(random_state)

        choices = rng.choice(
            samples, size=(min(ngen, samples.size),), replace=False
        )

        template_array = np.array(self.flattened())

        for choice in choices:
            template_array[chain_pars] = chains[..., choice]
            yield np.asarray(template_array).astype(float, copy=False)


class BaseParameter:
    __add__ = MAKE_BINARY(operator.add)
    __sub__ = MAKE_BINARY(operator.sub)
    __mul__ = MAKE_BINARY(operator.mul)
    __truediv__ = MAKE_BINARY(operator.truediv)
    __pow__ = MAKE_BINARY(operator.pow)
    __radd__ = MAKE_RBINARY(operator.add)
    __rsub__ = MAKE_RBINARY(operator.sub)
    __rmul__ = MAKE_RBINARY(operator.mul)
    __rtruediv__ = MAKE_RBINARY(operator.truediv)
    __abs__ = MAKE_UNARY(operator.abs)
    __mod__ = MAKE_BINARY(operator.mod)
    __rmod__ = MAKE_RBINARY(operator.mod)
    __lt__ = MAKE_BINARY(operator.lt)
    __le__ = MAKE_BINARY(operator.le)
    __ge__ = MAKE_BINARY(operator.ge)
    __gt__ = MAKE_BINARY(operator.gt)
    __neg__ = MAKE_UNARY(operator.neg)

    def __init__(self):
        # add mathematical operations as methods to this object

        for k, v in math_ops.items():
            setattr(self, k, MethodType(v, self))

        self._vary = False
        self.name = None
        self._value = None
        self._bounds = None
        self._deps = []
        self._stderr = None
        self.chain = None

    def __getstate__(self):
        # the mathematical ops aren't unpickleable, so lets pop them. They'll
        # be reinstated when the constructor is called anyway.

        # we need to retain the mathops in this object, but not in the pickle
        # dict. Perform a shallow copy.
        d = self.__dict__.copy()
        for k in ops:
            d.pop(k, None)
        return d

    def __setstate__(self, state):
        # the mathematical ops aren't unpickleable, so lets pop them. They'll
        # be reinstated when the constructor is called anyway.
        for k, v in math_ops.items():
            setattr(self, k, MethodType(v, self))

        self.__dict__.update(state)

    def __or__(self, other):
        # concatenate parameter with parameter or parameters
        if not (is_parameter(other) or is_parameters(other)):
            raise ValueError(
                "Can only concatenate a Parameter with another"
                " Parameter or Parameters instance"
            )

        p = Parameters()
        p.append(self)
        p.append(other)
        return p

    @property
    def vary(self):
        return self._vary

    def logp(self):
        raise NotImplementedError(
            "Subclass of BaseParameter should override this method"
        )

    def __float__(self):
        return float(self.value)

    def __bool__(self):
        return bool(self.value)

    def __numpy_ufunc__(self):
        return np.array(self.value)

    @property
    def value(self):
        return self._eval()

    @value.setter
    def value(self, v):
        self._value = v

    @property
    def constraint(self):
        return None

    @property
    def bounds(self):
        return None

    @property
    def stderr(self):
        return self._stderr

    @stderr.setter
    def stderr(self, val):
        try:
            self._stderr = float(val)
        except TypeError:
            self._stderr = None

    def dependencies(self):
        dep_list = []
        for _dep in self._deps:
            if isinstance(_dep, Parameter):
                if _dep.constraint is not None:
                    dep_list.append(_dep.dependencies())
                else:
                    dep_list.append(_dep)
            if isinstance(_dep, (_UnaryOp, _BinaryOp)):
                dep_list.append(_dep.dependencies())
        return list(f_unique(flatten(dep_list)))

    def has_dependencies(self):
        return len(self._deps) > 0

    def __str__(self):
        """Returns printable representation of a Parameter object."""
        constraint = ""
        if self.constraint is not None:
            constraint = f", constraint={self.constraint}"

        fixed = ""
        if not self.vary and self.constraint is None:
            fixed = " (fixed)"
        elif self.stderr is not None:
            fixed = f" +/- {self.stderr:0.3g}"

        s = (
            f"<Parameter:{self.name!r:^15}"
            f", value={self.value:g}{fixed: ^10}"
            f", bounds={str(self.bounds)}"
            f"{constraint}>"
        )
        return s


class Constant(BaseParameter):
    def __init__(self, value=0.0, name=None):
        super().__init__()
        self.name = name
        self.value = value
        self._vary = False

    def __repr__(self):
        return f"Constant(value={self.value}, name={self.name!r})"

    def _eval(self):
        return self._value


def possibly_create_parameter(
    value, name="", bounds=None, vary=False, constraint=None, units=None
):
    """
    If supplied with a Parameter return it. If supplied with float, wrap it in
    a Parameter instance.

    Parameters
    ----------
    value : float, optional
        Numerical Parameter value.
    name : str, optional
        Name of the parameter.
    bounds: `refnx.analysis.Bounds`, tuple, optional
        Sets the bounds for the parameter. Either supply a
        `refnx.analysis.Bounds` object (or one of its subclasses),
        or a `(lower_bound, upper_bound)` tuple.
    vary : bool, optional
        Whether the Parameter is fixed during a fit.
    constraint : expression, optional
        Python expression used to constrain the value during the fit.
    units : str
        Units for the Parameter.

    Returns
    -------
    parameter : refnx.analysis.Parameter

    """
    if is_parameter(value):
        return value
    else:
        return Parameter(
            value,
            name=name,
            bounds=bounds,
            vary=vary,
            constraint=constraint,
            units=units,
        )


def sequence_to_parameters(seq):
    """
    Flattens and converts sequence of float/Parameter to a Parameters instance.

    Parameters
    ----------
    seq: sequence
        Sequence (possibly nested) of float/Parameter

    Returns
    -------
    params: `refnx.analysis.Parameters`
    """
    flat_seq = flatten(seq)
    pars = [possibly_create_parameter(p) for p in flat_seq]
    return Parameters(pars)


class Parameter(BaseParameter):
    """
    Class for specifying a variable.

    Parameters
    ----------
    value : float, optional
        Numerical Parameter value.
    name : str, optional
        Name of the parameter.
    bounds: `refnx.analysis.Bounds`, tuple, optional
        Sets the bounds for the parameter. Either supply a
        `refnx.analysis.Bounds` object (or one of its subclasses),
        or a `(lower_bound, upper_bound)` tuple.
    vary : bool, optional
        Whether the Parameter is fixed during a fit.
    constraint : expression, optional
        Python expression used to constrain the value during the fit.
    """

    def __init__(
        self,
        value=0.0,
        name=None,
        bounds=None,
        vary=False,
        constraint=None,
        units=None,
    ):
        """
        Class for specifying a variable.

        Parameters
        ----------
        value : float, optional
            Numerical Parameter value.
        name : str, optional
            Name of the parameter.
        bounds: `refnx.analysis.Bounds`, tuple, optional
            Sets the bounds for the parameter. Either supply a
            `refnx.analysis.Bounds` object (or one of its subclasses),
            or a `(lower_bound, upper_bound)` tuple.
        vary : bool, optional
            Whether the Parameter is fixed during a fit.
        constraint : expression, optional
            Python expression used to constrain the value during the fit.
        units : str, optional
            units for the Parameter
        """
        super().__init__()

        self.name = name

        # set bounds before setting value because the value may not be valid
        # for the bounds
        self._bounds = Interval()
        if bounds is not None:
            self.bounds = bounds

        self.value = value
        self._vary = vary

        self._constraint = None
        self._constraint_args = None
        self.constraint = constraint
        self.units = units

    def __repr__(self):
        # repr does not include stderr because that can't be used to
        # create a Parameter
        s = (
            f"{self.__class__.__name__}("
            f"value={float(self.value)},"
            f" name={self.name!r},"
            f" vary={self.vary!r},"
            f" bounds={self._bounds!r},"
            f" constraint={self._constraint!r})"
        )
        return s

    def logp(self, pval=None):
        """
        Calculate the log probability of the parameter

        Returns
        -------
        prob : float
            log probability of the parameter
        """
        if pval is not None:
            val = pval
        else:
            val = self.value

        if self.bounds is not None:
            return self.bounds.logp(val)
        else:
            return 0.0

    @property
    def value(self):
        """
        The numeric value of the :class:`Parameter`
        """
        if self._constraint is not None:
            if callable(self._constraint):
                retval = float(self._constraint(*self._constraint_args))
            else:
                retval = self._constraint._eval()
        else:
            retval = self._value

        return retval

    @value.setter
    def value(self, v):
        value = np.float64(v)
        self._value = value

    def _eval(self):
        if self._constraint is not None:
            if isinstance(self._constraint, BaseParameter):
                return self._constraint._eval()
            elif callable(self._constraint):
                return self.value
        else:
            return self._value

    @property
    def bounds(self):
        """
        The bounds placed on this :class:`Parameter`.
        """
        return self._bounds

    @bounds.setter
    def bounds(self, bounds):
        if isinstance(bounds, Bounds):
            self._bounds = bounds
        elif bounds is None:
            self._bounds = Interval()
        elif hasattr(bounds, "__len__") and len(bounds) == 2:
            self.range(*bounds)
        else:
            rv = PDF(bounds)
            self._bounds = rv

    def valid(self, val):
        """
        The truth of whether a value would satisfy the bounds for this
        parameter.

        Parameters
        ----------
        val : float
            A proposed value

        Returns
        -------
        valid : bool
            `np.isfinite(Parameter.logp(val))`
        """
        return self.bounds.valid(val)

    def range(self, lower, upper):
        """
        Sets the lower and upper limits on the Parameter

        Parameters
        ----------
        lower : float
            lower bound
        upper : float
            upper bound

        Returns
        -------
        None

        """
        self.bounds = Interval(lower, upper)

    @property
    def vary(self):
        """
        Whether this :class:`Parameter` is allowed to vary
        """
        return self._vary

    @vary.setter
    def vary(self, vary):
        if self.constraint is not None:
            raise RuntimeError("cannot vary a Parameter which is constrained")
        else:
            self._vary = vary

    @property
    def constraint(self):
        return self._constraint

    @constraint.setter
    def constraint(self, expr):
        self._deps = []
        if expr is None:
            value = self.value
            self._constraint = None
            self._constraint_args = None
            self.value = value
            return
        _expr = asMagicNumber(expr)
        if id(self) in [id(dep) for dep in flatten(_expr.dependencies())]:
            raise ValueError("Your constraint contains a circular dependency")
        self._constraint = _expr
        self._constraint_args = None
        if isinstance(expr, Parameter):
            self._deps.append(expr)
        self._deps.extend(flatten(_expr.dependencies()))
        self._vary = False

    def set_constraint(self, constraint, args=()):
        """
        Constrains the Parameter.

        Parameters
        ----------
        constraint : {None, expression, callable}
            One of:

            - None, remove all constraints on this Parameter.
            - expression, an algebraic Python expression used to constrain the
               Parameter value.
            - callable, a Python function, ``constraint(*args)`` that returns
               a float value for the Parameter value.

            The callable should not use this Parameter in any of its
            calculations; nor should the callable use any Parameter in its
            calculations that possesses a constraint that would eventually
            lead back to this Parameter. If these conditions aren't met then
            circular dependencies with undefined side effects will be
            created. *A Parameter cannot ultimately depend on itself*.

        args : tuple
            a sequence of arguments given to `constraint` if it is a callable.
            This sequence can contain other Parameters, numbers, arrays,
            objects, etc. It is *strongly recommended* that this sequence is
            not nested. This is because ``args`` is searched for other
            Parameter objects, which are stored internally within this object
            as dependencies. Any constraints that these dependencies may have
            are evaluated before they are provided to the callable.
            If the callable uses Parameters that are not immediately
            retrievable from ``args`` (e.g. stored as attributes in an object),
            and those Parameters have constraints themselves, then those
            Parameters will likely have stale values, resulting in undefined
            behaviour.

        Examples
        --------
        >>> from refnx.analysis import Parameter
        >>> a = Parameter(1)
        >>> b = Parameter(2)
        >>> a.set_constraint(np.sin(2*b))
        >>> print(a.value)
        -0.7568024953079282

        >>> def c(*args):
        ...     return np.sin(args[0] * args[1])

        >>> a.set_constraint(c, args=(2, b))
        >>> print(a.value)
        -0.7568024953079282
        """
        if constraint is None:
            # clear the constraint
            self.constraint = None
        elif callable(constraint):
            deps = []
            for arg in flatten(args):
                if id(arg) == id(self):
                    raise ValueError(
                        "Your constraint contains a circular dependency"
                    )

                if isinstance(arg, BaseParameter):
                    expr = asMagicNumber(arg)
                    if id(self) in [
                        id(dep) for dep in flatten(expr.dependencies())
                    ]:
                        raise ValueError(
                            "Your constraint contains a circular dependency"
                        )
                    if isinstance(expr, Parameter):
                        deps.append(expr)
                    deps.extend(flatten(expr.dependencies()))

            # check return value can be coerced to a float
            # an Exception will probably be raised if that's the case
            float(constraint(*args))

            # at this point the constraint function should be ok.
            self._constraint = constraint
            self._constraint_args = args
            self._deps = deps
            self._vary = False
        else:
            self.constraint = constraint

    def setp(self, value=None, vary=None, bounds=None, constraint=None):
        """
        Set several attributes of the parameter at once.

        Parameters
        ----------
        value : float, optional
            Numerical Parameter value.
        vary : bool, optional
            Whether the Parameter is fixed during a fit.
        bounds: `refnx.analysis.Bounds`, tuple, optional
            Sets the bounds for the parameter. Either supply a
            `refnx.analysis.Bounds` object (or one of its subclasses),
            or a `(lower_bound, upper_bound)` tuple.
        constraint : expression, optional
            Python expression used to constrain the value during the fit.
        """
        if value is not None:
            self.value = value
        if vary is not None:
            self.vary = vary
        if bounds is not None:
            self.bounds = bounds
        if constraint is not None:
            self.constraint = constraint

    def corner(self):
        """
        Plots a histogram of the Parameter posterior.

        Requires matplotlib be installed.

        Returns
        -------
        fig, ax : :class:`matplotlib.Figure`, :class:`matplotlib.Axes`
            `matplotlib` figure and axes objects.
        """
        if self.chain is None:
            raise RuntimeError(
                "Parameter.chain is not present try sampling first"
            )
        import matplotlib.pyplot as plt

        fig = plt.figure()
        ax = fig.add_subplot(111)
        lb, med, ub = np.percentile(self.chain, [15.87, 50, 84.13])
        sigma = 0.5 * (ub - lb)

        s = ax.hist(
            self.chain.flatten(), bins=100, density=True, histtype="step"
        )
        ax.vlines(
            [med - sigma, med, med + sigma],
            0,
            1.1 * np.max(s[0]),
            colors="r",
            linestyle=["dashed", "solid", "dashed"],
        )
        ax.set_xlabel(f"{self.name} / {self.units}")
        ax.annotate(f"{med:.6g} $\\pm$ {sigma:.6g}", (med, np.max(s[0])))
        return fig, ax


class _BinaryOp(BaseParameter):
    def __init__(self, op1, op2, operation):
        super().__init__()
        self.op1 = op1
        self.op2 = op2
        self.opn = operation
        self._deps = []
        self._deps.append(op1)
        self._deps.append(op2)

    def _eval(self):
        return self.opn(self.op1._eval(), self.op2._eval())


class _UnaryOp(BaseParameter):
    def __init__(self, op1, operation):
        super().__init__()
        self.op1 = op1
        self.opn = operation
        self._deps = []
        self._deps.append(op1)

    def _eval(self):
        return self.opn(self.op1._eval())


def is_parameter(x):
    """Test for Parameter-ness."""
    return isinstance(x, BaseParameter)


def is_parameters(x):
    """Test for Parameter-ness."""
    return isinstance(x, Parameters)


def _constraint_tree_helper(expr):
    t = []
    if isinstance(expr, Parameter):
        return expr
    if isinstance(expr, Constant):
        return expr
    if isinstance(expr, _BinaryOp):
        t.append(
            [
                _constraint_tree_helper(expr.op1),
                _constraint_tree_helper(expr.op2),
                expr.opn,
            ]
        )
    if isinstance(expr, _UnaryOp):
        t.append([_constraint_tree_helper(expr.op1), expr.opn])
    return t


def constraint_tree(expr):
    """
    builds a mathematical tree of a constraint expression
    this can be fed into build_constraint_from_tree to
    reconstitute a constraint
    """
    if isinstance(expr, Parameter):
        return [expr]
    if isinstance(expr, Constant):
        return [expr]
    return list(flatten(_constraint_tree_helper(expr)))


def build_constraint_from_tree(tree):
    """
    A calculator for a constraint tree. It's essentially a reverse
    Polish notation calculator.
    """
    v = []
    for t in tree:
        if callable(t):
            if t in binary:
                o1 = v.pop()
                o2 = v.pop()
                v.append(t(o1, o2))
            elif t in unary:
                o1 = v.pop()
                v.append(t(o1))
        else:
            v.append(t)

    return v[0]



from importlib import resources
from pathlib import Path
import os
import pickle

import numpy as np

from refnx.reflect import SLD, Slab, ReflectModel, Motofit
from refnx.dataset import ReflectDataset
import refnx.reflect.tests

from numpy.testing import (
    assert_equal,
    assert_,
)


class Test__InteractiveModeller:
    def setup_method(self):
        self.pth = resources.files(refnx.reflect.tests)

        sio2 = SLD(3.47, name="SiO2")
        air = SLD(0, name="air")
        si = SLD(2.07, name="Si")
        d2o = SLD(6.36, name="D2O")
        polymer = SLD(1, name="polymer")

        self.structure = air | sio2(100, 2) | si(0, 3)

        theoretical = np.loadtxt(self.pth / "theoretical.txt")
        qvals, rvals = np.hsplit(theoretical, 2)
        self.qvals = qvals.flatten()
        self.rvals = rvals.flatten()

        # e361 is an older dataset, but well characterised
        self.structure361 = si | sio2(10, 4) | polymer(200, 3) | d2o(0, 3)
        self.model361 = ReflectModel(self.structure361, bkg=2e-5)

        self.model361.scale.vary = True
        self.model361.bkg.vary = True
        self.model361.scale.range(0.1, 2)
        self.model361.bkg.range(0, 5e-5)

        # d2o
        self.structure361[-1].sld.real.vary = True
        self.structure361[-1].sld.real.range(6, 6.36)

        self.structure361[1].thick.vary = True
        self.structure361[1].thick.range(5, 20)

        self.structure361[2].thick.vary = True
        self.structure361[2].thick.range(100, 220)

        self.structure361[2].sld.real.vary = True
        self.structure361[2].sld.real.range(0.2, 1.5)

        self.e361 = ReflectDataset(self.pth / "e361r.txt")
        self.qvals361, self.rvals361, self.evals361 = (
            self.e361.x,
            self.e361.y,
            self.e361.y_err,
        )
        self.app = Motofit()
        self.app(self.e361, model=self.model361)

    def test_run_app(self):
        # figure out it some of the parameters are the same as you set them
        # with
        slab_views = self.app.model_view.structure_view.slab_views

        for slab_view, slab in zip(slab_views, self.structure361):
            assert_equal(slab_view.w_thick.value, slab.thick.value)
            assert_equal(slab_view.w_sld.value, slab.sld.real.value)
            assert_equal(slab_view.w_isld.value, slab.sld.imag.value)
            assert_equal(slab_view.w_rough.value, slab.rough.value)

            assert_equal(slab_view.c_thick.value, slab.thick.vary)
            assert_equal(slab_view.c_sld.value, slab.sld.real.vary)
            assert_equal(slab_view.c_isld.value, slab.sld.imag.vary)
            assert_equal(slab_view.c_rough.value, slab.rough.vary)

    def test_fit_runs(self):
        parameters = self.app.objective.parameters
        assert_(self.app.dataset is not None)
        assert_(self.app.model.structure[1].thick.vary is True)

        var_params = len(parameters.varying_parameters())
        assert_equal(var_params, 6)

        # sometimes the output buffer gets detached.
        try:
            self.app.do_fit(None)
        except ValueError:
            pass



%%cython -a

# cython: language_level=3, cdivision=False
from collections import namedtuple
from refnx._lib import possibly_open_file
import numpy as np

from libc.stdint cimport uint32_t, int64_t, int32_t, uint8_t
cimport numpy as cnp
cimport cython

import mmap

ii32 = np.iinfo(np.int32)


EventFileHeader_Base = namedtuple('EventFileHeader_Base',
                                  ['magic_number', 'format_number',
                                   'anstohm_version', 'pack_format',
                                   'oob_enabled', 'clock_scale'])
EventFileHeader_Packed = namedtuple('EventFileHeader_Packed',
                                    ['evt_stg_nbits_x', 'evt_stg_nbits_y',
                                     'evt_stg_nbits_v', 'evt_stg_nbits_w',
                                      'evt_stg_nbits_wa', 'evt_stg_xy_signed'])

def event_header(f):
    """
    Reads the header from an ANSTO event file

    Parameters
    ----------
    f : str or file-like
        The event file of interest

    Returns
    -------
    base, packed : EventFileHeader_Base, EventFileHeader_Packed
    """
    with possibly_open_file(f, 'rb') as fh:
        fh.seek(0)
        header = fh.read(128)
        header_arr = np.frombuffer(header, dtype='int32')

    base = EventFileHeader_Base(magic_number=header_arr[0],
                                format_number=header_arr[1],
                                anstohm_version = header_arr[2],
                                pack_format = header_arr[3],
                                oob_enabled = header_arr[4],
                                clock_scale = header_arr[5])
    assert(base.magic_number == 0x0DAE0DAE)

    if header_arr[1] >= 0x00010002:
        evt_stg_nbits_wa = header_arr[20]
        evt_stg_xy_signed = header_arr[21]
    else:
        evt_stg_nbits_wa = 0
        evt_stg_xy_signed = header_arr[20]

    packed = EventFileHeader_Packed(evt_stg_nbits_x=header_arr[16],
                            evt_stg_nbits_y=header_arr[17],
                            evt_stg_nbits_v=header_arr[18],
                            evt_stg_nbits_w=header_arr[19],
                            evt_stg_nbits_wa=evt_stg_nbits_wa,
                            evt_stg_xy_signed=evt_stg_xy_signed)
    return base, packed


"""
event decoding state machine
for all events
initial state - then DECODE_VAL_BITFIELDS (for neutron
events) or DECODE_OOB_BYTE_1 (for OOB events) for OOB events only
"""
cdef int DECODE_START = 0
cdef int DECODE_OOB_BYTE_1 = 1
cdef int DECODE_OOB_BYTE_2 = 2
# for all events
cdef int DECODE_VAL_BITFIELDS = 3
# final state - then output data and return to DECODE_START
cdef int DECODE_DT = 4

# all events contain some or all of these fields
# x, y, v, w, wa
DEF NVAL = 5

"""
Types of OOB events, and 'NEUTRON' event.  Not all are used for all
instruments, or supported yet.

NEUTRON = 0 = a neutron detected, FRAME_START = -2 = T0 pulse (e.g. from
chopper, or from Doppler on Emu).  For most instruments, these are the only
types used.

FRAME_AUX_START = -3 (e.g. from reflecting chopper on Emu), VETO = -6 (e.g.
veto signal from ancillary)

BEAM_MONITOR = -7 (e.g. if beam monitors connected direct to Mesytec MCPD8 DAE)

RAW = -8 = pass-through, non-decoded raw event directly from the DAE (e.g.
Mesytec MCPD8).  Used to access special features of DAE.

Other types are not used in general (DATASIZES = -1 TBD in future, FLUSH = -4
deprecated, FRAME_DEASSERT = -5 only on Fastcomtec P7888 DAE).
"""

@cython.boundscheck(False)
@cython.cdivision(False)
def events(f, def_clock_scale=1000, use_tx_chopper=False):

    # read file header (base header then packed-format header)
    hdr_base, hdr_packed = event_header(f)

    if hdr_base.pack_format:
        raise RuntimeError("only packed binary format is supported")

    # data starts at byte 128
    f.seek(128)

    """
    Setup the clock_scale.  In format 0x00010001 this was not part of the
    headers, hence a function argument is provided to allow it to be
    specified manually. In the current format 0x00010002, clock_scale is
    written to the header and need not be specified, unless some alternate
    scale is needed.
    """
    scale_microsec = 1 / hdr_base.clock_scale
    if not hdr_base.clock_scale:
        # old eventfile format did not have clock_scale...
        scale_microsec = 1 / def_clock_scale

    # the initial time is not set correctly so wait until primary and auxillary
    # time have been reset before sending events
    cdef:
        int64_t primary_time = 0
        int64_t auxillary_time = 0
        bint primary_ok = False
        bint auxillary_ok = False
        bint evt_stg_xy_signed = hdr_packed.evt_stg_xy_signed

    cdef:
        # event data fields
        uint32_t x = 0, y = 0, v = 0, w = 0, wa = 0
        uint32_t ptr_val[NVAL]
        Py_ssize_t frame_number = -1
        int32_t signed_x, signed_y

    ptr_val = [0, 0, 0, 0, 0]

    """
    All events are also timestamped.  The differential timestamp dt stored in
    each event is summed to recover the event timestamp t. All timestamps are
    frame-relative, i.e. FRAME_START event represents T0 (e.g. from a chopper)
    and t is reset to 0. In OOB mode and for certain DAE types only (e.g.
    Mesytec MCPD8), the FRAME_START event is timestamped relative to the last
    FRAME_START. The timestamp t on the FRAME_START event is therefore the
    total frame duration, and this can be used to recover the absolute
    timestamp of all events in the DAQ, if desired (e.g. for accurate timing
    during long term kinematic experiments).
    """
    # t = 0 dt may be negative occasionally for some DAE types, therefore dt
    # and t are signed ints.
    cdef:
        int32_t dt
        int32_t ind_val = 0
        int32_t nbits_val = 0
        int32_t nbits_val_filled = 0
        int32_t nbits_dt_filled = 0
        int32_t nbits_val_to_fill
        int32_t nbits_ch_used = 0
        uint8_t ch
        bint frame_start_event
        bint event_ended = False
        int32_t oob_en
        int32_t oob_event = 0
        int32_t c = 0
        int32_t nbits_val_neutron[NVAL]
        int32_t nbits_val_oob[NVAL]
        int state = DECODE_START
        int num_events = 0

    nbits_val_oob = [0, 0, 0, 0, 0]
    nbits_val_neutron = [
        hdr_packed.evt_stg_nbits_x,
        hdr_packed.evt_stg_nbits_y,
        hdr_packed.evt_stg_nbits_v,
        hdr_packed.evt_stg_nbits_w,
        hdr_packed.evt_stg_nbits_wa]

    # will be 1 if we are reading a new OOB
    # event file (format 0x00010002 only).
    oob_en = hdr_base.oob_enabled

    # For neutron events, oob_event = 0, and for OOB events,
    # oob_event = 1 and c indicates the OOB event type.
    # c < 0 for all OOB events currently.
    oob_event = 0
    c = 0

    # event decoding state machine
    state = DECODE_START

    while True:
        # read next byte
        www = f.read(1)
        if len(www):
            ch = www[0]
        else:
            break
        # no bits used initially, 8 to go
        nbits_ch_used = 0

        # start of event processing
        if (state == DECODE_START):
            # if OOB event mode is enabled, the leading Bit 0 of the first byte
            # indicates whether the event is a neutron event or an OOB event
            if not oob_en:
                state = DECODE_VAL_BITFIELDS
            else:
                oob_event = ch & 1
                # leading bit used as OOB bit
                nbits_ch_used = 1

                if not oob_event:
                    state = DECODE_VAL_BITFIELDS
                else:
                    state = DECODE_OOB_BYTE_1

            # setup to decode new event bitfields (for both neutron
            # and OOB events)
            for ind_val in range(NVAL):
                ptr_val[ind_val] = 0

            ind_val = 0
            nbits_val_filled = 0

            dt = 0
            nbits_dt_filled = 0

        # state machine for event decoding
        if state == DECODE_START:
            raise RuntimeError('Failure in event decoding')
        elif state == DECODE_OOB_BYTE_1:
            """
            first OOB header byte
            OOB event Byte 1:  Bit 0 = 1 = OOB event, Bit 1 =
            mode (only mode=0 suported currently), Bits 2-5 =
            c (OOB event type), Bits 6-7 = bitfieldsize_x
            / 8. bitfieldsize_x and following 2-bit
            bitfieldsizes are the number of bytes used to
            store the OOB parameter. All of x,y,v,w,wa are
            short integers (16 bits maximum) and so
            bitfieldsizes = 0, 1 or 2 only.
            """
            # Bits 2-5 = c
            c = ch >> 2 & 0xF

            if c & 0x8:
                # c is a signed parameter so sign extend - OOB
                # events are negative values
                c |= <int32_t> 0xFFFFFFF0
            nbits_val_oob[0] = (ch & 0xC0) >> 3  # Bits 6-7 * 8 = bitfieldsize_x

            # Proceed to process second OOB event header
            # byte next time
            state = DECODE_OOB_BYTE_2
        elif state == DECODE_OOB_BYTE_2:
            # second OOB header byte
            # bitfieldsizes for y, v, w and wa, as for
            # bitfieldsize_x above.
            nbits_val_oob[1] = (ch & 0x03) << 3  # Bits 0-1 * 8 = bitfieldsize_y
            nbits_val_oob[2] = (ch & 0x0C) << 1  # Bits 2-3 * 8 = bitfieldsize_v
            nbits_val_oob[3] = (ch & 0x30) >> 1  # Bits 4-5 * 8 = bitfieldsize_w
            nbits_val_oob[4] = (ch & 0xC0) >> 3  # Bits 6-7 * 8 = bitfieldsize_wa

            # Proceed to read and store x, y, v, w, wa for
            # the OOB event
            state = DECODE_VAL_BITFIELDS
        elif state == DECODE_VAL_BITFIELDS:
            # fill bits of the incoming ch to the event's bitfields.
            # stop when we've filled them all, or all bits of ch are used.
            while (ind_val < NVAL) and (nbits_ch_used < 8):
                nbits_val = nbits_val_oob[ind_val]
                if not oob_event:
                    nbits_val = nbits_val_neutron[ind_val]
                if not nbits_val:
                    nbits_val_filled = 0
                    ind_val += 1
                else:
                    nbits_val_to_fill = nbits_val - nbits_val_filled
                    if (8 - nbits_ch_used) >= nbits_val_to_fill:
                        ptr_val[ind_val] |= (((ch >> nbits_ch_used) &
                                              ((1 << nbits_val_to_fill) - 1))
                                               << nbits_val_filled)
                        nbits_val_filled = 0
                        nbits_ch_used += nbits_val_to_fill
                        ind_val += 1
                    else:
                        ptr_val[ind_val] |= (ch >> nbits_ch_used) << nbits_val_filled
                        nbits_val_filled += (8 - nbits_ch_used)
                        nbits_ch_used = 8

            if ind_val == NVAL:
                # and fall through for dt processing
                state = DECODE_DT

            if nbits_ch_used == 8:
                # read next byte
                continue

        if state == DECODE_DT:
            if 8 - nbits_ch_used <= 2:
                dt |= (ch >> nbits_ch_used) << nbits_dt_filled
                nbits_dt_filled += 8 - nbits_ch_used
            elif (ch & 0xC0) == 0xC0:
                dt |= ((ch & 0x3F) >> nbits_ch_used) << nbits_dt_filled
                nbits_dt_filled += (6 - nbits_ch_used)
            else:
                dt |= (ch >> nbits_ch_used) << nbits_dt_filled
                nbits_dt_filled += (8 - nbits_ch_used)
                event_ended = True

        # unpack values from array
        # this would be better done by using a pointer array
        x, y, v, w, wa = ptr_val

        if event_ended:
            # start on new event next time
            state = DECODE_START
            # update times
            primary_time += dt
            auxillary_time += dt

            # is this event a frame_start?
            # FRAME_START is an OOB event when oob
            # mode enabled
            if oob_en:
                frame_start_event = oob_event and c == -2
            else:
                frame_start_event = (x == 0 and y == 0 and dt == -1)

            if oob_en or (not frame_start_event):
                if oob_event:
                    if c == -3:
                        # FRAME_AUX_START = -3
                        # 0 is the reflecting chopper and 1 is the transmission chopper
                        if not use_tx_chopper and x == 0:
                            auxillary_time = 0
                            auxillary_ok = True
                        if use_tx_chopper and x == 1:
                            auxillary_time = 0
                            auxillary_ok = True
                else:
                    # if times are ok, time units in usec
                    if primary_ok and auxillary_ok:
                        if evt_stg_xy_signed:
                            # if x and y are signed then convert from uint32 to int32
                            signed_x, signed_y = x, y
                            if x & (2**(nbits_val_neutron[0] - 1)):
                                signed_x = - (<uint32_t>0x100000000 -
                                              (x | <uint32_t>0xFFFFFC00))
                            if y & (2**(nbits_val_neutron[1] - 1)):
                                signed_y = - (<uint32_t>0x100000000 -
                                              (y | <uint32_t>0xFFFFFC00))
                        # TODO ADD AN EVENT TO A LIST SOMEWHERE
                        print(signed_x, signed_y, primary_time * scale_microsec)
                        num_events += 1

            if frame_start_event:
                # reset timestamp at start of a new frame
                # the auxillary time is only available in OOB mode
                # otherwise, auxillary time = primary time
                primary_time = 0
                primary_ok = True
                if not oob_en:
                    auxillary_time = 0
                    auxillary_ok = True

                # TODO INCREMENT A FRAME NUMBER
                frame_number += 1

            # TODO UPDATE FILE POSITION
            # progress.update(loader.selected_position())

            event_ended = False
    return num_events, frame_number



import multiprocessing
import warnings
import sys


if __name__ == "__main__":
    # needed for windows using multiprocessing!
    multiprocessing.freeze_support()
    warnings.filterwarnings("ignore", "(?s).*MATPLOTLIBDATA.*",
                            category=UserWarning)
    from refnx.reflect import main
    main()



#!/bin/bash

"""
Using refnx in a highly parallelised environment using mpi.

You'll need to install:
- refnx
- numpy
- cython
- schwimmbad
- mpi4py

Usage
-----
mpiexec -n 4 python mpi_parallelisation.py
"""

# Start off by importing necessary packages
import sys
import os.path

import refnx
from schwimmbad import MPIPool

from refnx.reflect import SLD, Slab, ReflectModel
from refnx.dataset import ReflectDataset
from refnx.analysis import (Objective, CurveFitter, Transform, GlobalObjective)


def setup():
    # load the data.
    DATASET_NAME = os.path.join(refnx.__path__[0],
                                'analysis',
                                'test',
                                'c_PLP0011859_q.txt')

    # load the data
    data = ReflectDataset(DATASET_NAME)

    # the materials we're using
    si = SLD(2.07, name='Si')
    sio2 = SLD(3.47, name='SiO2')
    film = SLD(2, name='film')
    d2o = SLD(6.36, name='d2o')

    structure = si | sio2(30, 3) | film(250, 3) | d2o(0, 3)
    structure[1].thick.setp(vary=True, bounds=(15., 50.))
    structure[1].rough.setp(vary=True, bounds=(1., 6.))
    structure[2].thick.setp(vary=True, bounds=(200, 300))
    structure[2].sld.real.setp(vary=True, bounds=(0.1, 3))
    structure[2].rough.setp(vary=True, bounds=(1, 6))

    model = ReflectModel(structure, bkg=9e-6, scale=1.)
    model.bkg.setp(vary=True, bounds=(1e-8, 1e-5))
    model.scale.setp(vary=True, bounds=(0.9, 1.1))
    model.threads = 1
    # fit on a logR scale, but use weighting
    objective = Objective(model, data, transform=Transform('logY'),
                          use_weights=True)

    return objective


def structure_plot(obj, samples=0):
    # plot sld profiles
    import matplotlib.pyplot as plt
    fig = plt.figure()
    ax = fig.add_subplot(111)

    if isinstance(obj, GlobalObjective):
        if samples > 0:
            savedparams = np.array(obj.parameters)
            for pvec in obj.parameters.pgen(ngen=samples):
                obj.setp(pvec)
                for o in obj.objectives:
                    if hasattr(o.model, 'structure'):
                        ax.plot(*o.model.structure.sld_profile(),
                                color="k", alpha=0.01)

            # put back saved_params
            obj.setp(savedparams)

        for o in obj.objectives:
            if hasattr(o.model, 'structure'):
                ax.plot(*o.model.structure.sld_profile(), zorder=20)

        ax.set_ylabel('SLD / $10^{-6}\\AA^{-2}$')
        ax.set_xlabel("z / $\\AA$")

    elif isinstance(obj, Objective) and hasattr(obj.model, 'structure'):
        fig, ax = obj.model.structure.plot(samples=samples)

    fig.savefig('steps_sld.png', dpi=1000)


if __name__ == "__main__":
    with MPIPool() as pool:
        if not pool.is_master():
            pool.wait()
            sys.exit(0)
        # buffering so the program doesn't try to write to the file
        # constantly
        with open('steps.chain', 'w', buffering=500000) as f:
            objective = setup()
            # Create the fitter and fit
            fitter = CurveFitter(objective, nwalkers=300)
            fitter.initialise('prior')
            fitter.fit('differential_evolution')
            # thin by 10 so we have a smaller filesize
            fitter.sample(100, pool=pool.map, f=f, verbose=False, nthin=10);
            f.flush()

        try:
            # create graphs of reflectivity and SLD profiles
            import matplotlib
            import matplotlib.pyplot as plt
            matplotlib.use('agg')

            fig, ax = objective.plot(samples=1000)
            ax.set_ylabel('R')
            ax.set_xlabel("Q / $\\AA$")
            fig.savefig('steps.png', dpi=1000)

            structure_plot(objective, samples=1000)

            # corner plot
            fig = objective.corner()
            fig.savefig('steps_corner.png')

            # plot the Autocorrelation function of the chain
            fig = plt.figure()
            ax = fig.add_subplot(111)
            ax.plot(fitter.acf())
            ax.set_ylabel('autocorrelation')
            ax.set_xlabel('step')
            fig.savefig('steps-autocorrelation.png')
        except ImportError:
            pass



"""
refnx: Neutron and X-ray reflectometry analysis in Python
=========================================================
Documentation is available in the docstrings and
online at https://readthedocs.org/projects/refnx/
"""

import importlib as _importlib

try:
    from refnx.version import version as __version__
except ImportError:
    __version__ = "version string not created yet"

from refnx._lib._testutils import PytestTester

test = PytestTester(__name__)
del PytestTester


submodules = [
    "analysis",
    "dataset",
    "reduce",
    "reflect",
    "util",
]

__all__ = submodules + [
    "test",
    "__version__",
]


def __dir__():
    return __all__


def __getattr__(name):
    if name in submodules:
        return _importlib.import_module(f"refnx.{name}")
    else:
        try:
            return globals()[name]
        except KeyError:
            raise AttributeError(f"Module 'refnx' has no attribute '{name}'")



import sys
import time
import os
import psutil
import subprocess


if __name__ == "__main__":
    # provide the path to the executable that you wish to check
    # print(sys.argv[1:])
    try:
        proc = subprocess.Popen(sys.argv[1:])
    except subprocess.TimeoutExpired:
        pass
    # leave time to start up
    pid = proc.pid
    time.sleep(10.0)

    l = (p.pid for p in psutil.process_iter())

    if pid not in l:
        raise RuntimeError("Motofit did not seem to be able to start.")
    else:
        # program started
        proc.kill()



"""
Component for studying lipid membranes at an interface
"""

import warnings
import numpy as np
from scipy.optimize import NonlinearConstraint
from refnx.reflect import Component, SLD, ReflectModel, Structure
from refnx.analysis import possibly_create_parameter, Parameters, Parameter
from refnx.reflect.structure import (
    overall_sld,
    Scatterer,
    possibly_create_scatterer,
)


class LipidLeaflet(Component):
    r"""
    Describes a lipid leaflet Component at an interface

    Parameters
    ----------
    APM: float or refnx.analysis.Parameter
    b_heads: float, refnx.analysis.Parameter, complex or SLD
        Sum of coherent scattering lengths of head group (Angstrom).
        When an SLD is provided it is simply an easy way to provide a complex
        value. `LipidLeaflet.b_heads_real` is set to `SLD.real`, etc.
    vm_heads: float or refnx.analysis.Parameter
        Molecular volume of head group (Angstrom**2)
    thickness_heads: float or refnx.analysis.Parameter
        Thickness of head group region (Angstrom)
    b_tails: float, refnx.analysis.Parameter, complex or SLD
        Sum of coherent scattering lengths of tail group (Angstrom).
        When an SLD is provided it is simply an easy way to provide a complex
        value. `LipidLeaflet.b_tails_real` is set to `SLD.real`, etc.
    vm_tails: float or refnx.analysis.Parameter
        Molecular volume of tail group (Angstrom**2)
    thickness_tails: float or refnx.analysis.Parameter
        Thickness of head group region (Angstrom)
    rough_head_tail: float or refnx.analysis.Parameter
        Roughness of head-tail group (Angstrom)
    rough_preceding_mono: float or refnx.analysis.Parameter
        Roughness between preceding component (in the fronting direction) and
        the monolayer (Angstrom). If `reverse_monolayer is False` then this is
        the roughness between the preceding component and the heads, if
        `reverse_monolayer is True` then this is the roughness between the
        preceding component and the tails.
    head_solvent: None, float, complex, refnx.reflect.SLD
        Solvent for the head region. If `None`, then solvation will be
        performed by the parent `Structure`, using the `Structure.solvent`
        attribute. Other options are coerced to an `SLD` object using
        `SLD(float | complex)`. A float/complex argument is the SLD of the
        solvent (10**-6 Angstrom**-2).
    tail_solvent: None, float, complex, refnx.reflect.SLD
        Solvent for the tail region. If `None`, then solvation will be
        performed by the parent `Structure`, using the `Structure.solvent`
        attribute. Other options are coerced to an `SLD` object using
        `SLD(float | complex)`. A float/complex argument is the SLD of the
        solvent (10**-6 Angstrom**-2).
    reverse_monolayer: bool, optional
        The default is to have heads closer to the fronting medium and
        tails closer to the backing medium. If `reverse_monolayer is True`
        then the tails will be closer to the fronting medium and heads
        closer to the backing medium.
    name: str, optional
        The name for the component

    Notes
    -----
    The sum of coherent scattering lengths must be in Angstroms, the volume
    must be in cubic Angstroms. This is because the SLD of a tail group is
    calculated as `b_tails / vm_tails * 1e6` to achieve the units
    10**6 Angstrom**-2.
    """

    # TODO: use SLD of head instead of b_heads, vm_heads?
    def __init__(
        self,
        apm,
        b_heads,
        vm_heads,
        thickness_heads,
        b_tails,
        vm_tails,
        thickness_tails,
        rough_head_tail,
        rough_preceding_mono,
        head_solvent=None,
        tail_solvent=None,
        reverse_monolayer=False,
        name="",
    ):
        """
        Parameters
        ----------
        apm: float or Parameter
            Area per molecule
        b_heads: float, Parameter or complex
            Sum of coherent scattering lengths of head group (Angstrom)
        vm_heads: float or Parameter
            Molecular volume of head group (Angstrom**3)
        thickness_heads: float or Parameter
            Thickness of head group region (Angstrom)
        b_tails: float, Parameter or complex
            Sum of coherent scattering lengths of tail group (Angstrom)
        vm_tails: float or Parameter
            Molecular volume of tail group (Angstrom**3)
        thickness_tails: float or Parameter
            Thickness of head group region (Angstrom)
        rough_head_tail: float or refnx.analysis.Parameter
            Roughness of head-tail group (Angstrom)
        rough_preceding_mono: float or Parameter
            Roughness between preceding component (in the fronting direction)
            and the monolayer (Angstrom). If `reverse_monolayer is False` then
            this is the roughness between the preceding component and the
            heads, if `reverse_monolayer is True` then this is the roughness
            between the preceding component and the tails.
        head_solvent: None, float, complex, SLD
            Solvent for the head region. If `None`, then solvation will be
            performed by the parent `Structure`, using the `Structure.solvent`
            attribute. Other options are coerced to an `SLD` object using
            `SLD(float | complex)`. A float/complex argument is the SLD of the
            solvent (10**-6 Angstrom**-2).
        tail_solvent: None, float, complex, SLD
            Solvent for the tail region. If `None`, then solvation will be
            performed by the parent `Structure`, using the `Structure.solvent`
            attribute. Other options are coerced to an `SLD` object using
            `SLD(float | complex)`. A float/complex argument is the SLD of the
            solvent (10**-6 Angstrom**-2).
        reverse_monolayer: bool, optional
            The default is to have heads closer to the fronting medium and
            tails closer to the backing medium. If `reverse_monolayer is True`
            then the tails will be closer to the fronting medium and heads
            closer to the backing medium.
        name: str, optional
            The name for the component
        """
        super().__init__()
        self.apm = possibly_create_parameter(
            apm, "%s - area_per_molecule" % name, units="Å**2"
        )

        if isinstance(b_heads, complex):
            self.b_heads_real = possibly_create_parameter(
                b_heads.real, name="%s - b_heads_real" % name
            )
            self.b_heads_imag = possibly_create_parameter(
                b_heads.imag, name="%s - b_heads_imag" % name
            )
        elif isinstance(b_heads, SLD):
            self.b_heads_real = b_heads.real
            self.b_heads_imag = b_heads.imag
        else:
            self.b_heads_real = possibly_create_parameter(
                b_heads, name="%s - b_heads_real" % name
            )
            self.b_heads_imag = possibly_create_parameter(
                0, name="%s - b_heads_imag" % name
            )

        self.b_heads_real.units = self.b_heads_imag.units = "Å"

        self.vm_heads = possibly_create_parameter(
            vm_heads, name="%s - vm_heads" % name, units="Å**3"
        )

        self.thickness_heads = possibly_create_parameter(
            thickness_heads, name="%s - thickness_heads" % name, units="Å"
        )

        if isinstance(b_tails, complex):
            self.b_tails_real = possibly_create_parameter(
                b_tails.real, name="%s - b_tails_real" % name
            )
            self.b_tails_imag = possibly_create_parameter(
                b_tails.imag, name="%s - b_tails_imag" % name
            )
        elif isinstance(b_tails, SLD):
            self.b_tails_real = b_tails.real
            self.b_tails_imag = b_tails.imag
        else:
            self.b_tails_real = possibly_create_parameter(
                b_tails, name="%s - b_tails_real" % name
            )
            self.b_tails_imag = possibly_create_parameter(
                0, name="%s - b_tails_imag" % name
            )
        self.b_tails_real.units = self.b_tails_imag.units = "Å"

        self.vm_tails = possibly_create_parameter(
            vm_tails, name="%s - vm_tails" % name, units="Å**3"
        )
        self.thickness_tails = possibly_create_parameter(
            thickness_tails, name="%s - thickness_tails" % name, units="Å"
        )
        self.rough_head_tail = possibly_create_parameter(
            rough_head_tail, name="%s - rough_head_tail" % name, units="Å"
        )
        self.rough_preceding_mono = possibly_create_parameter(
            rough_preceding_mono,
            name="%s - rough_fronting_mono" % name,
            units="Å",
        )

        self.head_solvent = self.tail_solvent = None
        if head_solvent is not None:
            self.head_solvent = possibly_create_scatterer(head_solvent)
        if tail_solvent is not None:
            self.tail_solvent = possibly_create_scatterer(tail_solvent)

        self.reverse_monolayer = reverse_monolayer
        self.name = name

    def __repr__(self):
        sld_bh = SLD([self.b_heads_real, self.b_heads_imag])
        sld_bt = SLD([self.b_tails_real, self.b_tails_imag])
        s = (
            f"LipidLeaflet("
            f"{self.apm!r}, "
            f"{sld_bh!r}, "
            f"{self.vm_heads!r}, "
            f"{self.thickness_heads!r}, "
            f"{sld_bt!r}, "
            f"{self.vm_tails!r}, "
            f"{self.thickness_tails!r}, "
            f"{self.rough_head_tail!r}, "
            f"{self.rough_preceding_mono!r}, "
            f"head_solvent={self.head_solvent!r}, "
            f"tail_solvent={self.tail_solvent!r}, "
            f"reverse_monolayer={self.reverse_monolayer}, "
            f"name={self.name!r})"
        )
        return s

    def slabs(self, structure=None):
        """
        Slab representation of monolayer, as an array

        Parameters
        ----------
        structure : refnx.reflect.Structure
            The Structure hosting this Component
        """
        layers = np.zeros((2, 5))

        # thicknesses
        layers[0, 0] = float(self.thickness_heads)
        layers[1, 0] = float(self.thickness_tails)

        # real and imag SLD's
        layers[0, 1] = float(self.b_heads_real) / float(self.vm_heads) * 1.0e6
        layers[0, 2] = float(self.b_heads_imag) / float(self.vm_heads) * 1.0e6

        layers[1, 1] = float(self.b_tails_real) / float(self.vm_tails) * 1.0e6
        layers[1, 2] = float(self.b_tails_imag) / float(self.vm_tails) * 1.0e6

        # roughnesses
        layers[0, 3] = float(self.rough_preceding_mono)
        layers[1, 3] = float(self.rough_head_tail)

        # volume fractions
        # head region
        layers[0, 4] = 1 - self.volfrac_h

        if self.head_solvent is not None:
            _head_solvent = self.head_solvent.complex(
                getattr(structure, "wavelength", None)
            )
            # we do the solvation here, not in Structure.slabs
            layers[0] = overall_sld(layers[0], _head_solvent)
            layers[0, 4] = 0

        # tail region
        layers[1, 4] = 1 - self.volfrac_t
        if self.tail_solvent is not None:
            _tail_solvent = self.tail_solvent.complex(
                getattr(structure, "wavelength", None)
            )
            # we do the solvation here, not in Structure.slabs
            layers[1] = overall_sld(layers[1], _tail_solvent)
            layers[1, 4] = 0

        if self.reverse_monolayer:
            layers = np.flipud(layers)
            layers[:, 3] = layers[::-1, 3]

        return layers

    @property
    def parameters(self):
        p = Parameters(name=self.name)
        p.extend(
            [
                self.apm,
                self.b_heads_real,
                self.b_heads_imag,
                self.vm_heads,
                self.thickness_heads,
                self.b_tails_real,
                self.b_tails_imag,
                self.vm_tails,
                self.thickness_tails,
                self.rough_head_tail,
                self.rough_preceding_mono,
            ]
        )
        if self.head_solvent is not None:
            p.append(self.head_solvent.parameters)
        if self.tail_solvent is not None:
            p.append(self.tail_solvent.parameters)

        return p

    def logp(self):
        # penalise unphysical volume fractions.
        if self.volfrac_h > 1 or self.volfrac_t > 1:
            return -np.inf

        return 0

    @property
    def volfrac_h(self):
        # Volume fraction of head group in head group region
        return self.vm_heads.value / (
            self.apm.value * self.thickness_heads.value
        )

    @property
    def volfrac_t(self):
        # Volume fraction of tail group in tail group region
        return self.vm_tails.value / (
            self.apm.value * self.thickness_tails.value
        )

    def make_constraint(self, objective):
        """
        Creates a NonlinearConstraint for a LipidLeaflet, ensuring that volume
        fraction of lipid in the head+tail regions lies in [0, 1]. Suitable for
        use by differential_evolution.

        Parameters
        ----------
        objective: refnx.analysis.Objective
            Objective containing the LipidLeaflet. Must be the Objective that is
            being minimised by differential_evolution.

        Returns
        -------
        nlc: NonlinearConstraint

        Notes
        -----
        You must create separate constraints for each LipidLeaflet object in your
        system.
        The Objective you supply must be for the overall curve fitting system.
        i.e. possibly a GlobalObjective.

        Examples
        --------
        >>> # leaflet is a LipidLeaflet, used in an Objective, obj
        >>> con = leaflet.make_constraint(obj)
        >>> fitter = CurveFitter(obj)
        >>> fitter.fit("differential_evolution", constraints=(con,))
        """

        def con(x):
            objective.setp(x)
            return self.volfrac_h, self.volfrac_t

        return NonlinearConstraint(con, 0, 1)


class LipidLeafletGuest(LipidLeaflet):
    r"""
    Describes a lipid leaflet Component at an interface with an incorporated guest molecule.

    Parameters
    ----------
    APM: float or refnx.analysis.Parameter
    b_heads: float, refnx.analysis.Parameter, complex or SLD
        Sum of coherent scattering lengths of head group (Angstrom).
        When an SLD is provided it is simply an easy way to provide a complex
        value. `LipidLeaflet.b_heads_real` is set to `SLD.real`, etc.
    vm_heads: float or refnx.analysis.Parameter
        Molecular volume of head group (Angstrom**2)
    thickness_heads: float or refnx.analysis.Parameter
        Thickness of head group region (Angstrom)
    b_tails: float, refnx.analysis.Parameter, complex or SLD
        Sum of coherent scattering lengths of tail group (Angstrom).
        When an SLD is provided it is simply an easy way to provide a complex
        value. `LipidLeaflet.b_tails_real` is set to `SLD.real`, etc.
    vm_tails: float or refnx.analysis.Parameter
        Molecular volume of tail group (Angstrom**2)
    thickness_tails: float or refnx.analysis.Parameter
        Thickness of head group region (Angstrom)
    rough_head_tail: float or refnx.analysis.Parameter
        Roughness of head-tail group (Angstrom)
    rough_preceding_mono: float or refnx.analysis.Parameter
        Roughness between preceding component (in the fronting direction) and
        the monolayer (Angstrom). If `reverse_monolayer is False` then this is
        the roughness between the preceding component and the heads, if
        `reverse_monolayer is True` then this is the roughness between the
        preceding component and the tails.
    phi_guest_h: float or refnx.analysis.Parameter
        Guest lying in the head layer. This is a fractional
        value representing how much of the space **not** taken up by the lipid
        is occupied by the guest molecule.
        If, however, `absolute_phi` is set to True, then this parameter
        represents the absolute value of the guest molecule in the head region.
        The absolute volume fraction is always available from the
        `LipidLeafletGuest.volfrac_guest_h` property.
        Please see the notes section for further details.

        .. warning::
           This parameter may not be determinable with low uncertainty if the lipid
           occupies nearly all of the space in the layer. For best results
           the guest and tail region should have very different SLDs.

    phi_guest_t: float or refnx.analysis.Parameter
        Guest lying in the tail layer. This is a fractional
        value representing how much of the space **not** taken up by the lipid
        is occupied by the guest molecule.
        If, however, `absolute_phi` is set to True, then this parameter
        represents the absolute value of the guest molecule in the tail region.
        The absolute volume fraction is always
        available from the `LipidLeafletGuest.volfrac_guest_t` property.
        Please see the notes section for further details.

        .. warning::
           This parameter may not be determinable with low uncertainty if the lipid
           occupies nearly all of the space in the layer. For best results
           the guest and tail region should have very different SLDs.

    sld_guest: None, float, complex, refnx.reflect.SLD
        SLD of guest molecule.
    head_solvent: None, float, complex, refnx.reflect.SLD
        Solvent for the head region. If `None`, then solvation will be
        performed by the parent `Structure`, using the `Structure.solvent`
        attribute. Other options are coerced to an `SLD` object using
        `SLD(float | complex)`. A float/complex argument is the SLD of the
        solvent (10**-6 Angstrom**-2).
    tail_solvent: None, float, complex, refnx.reflect.SLD
        Solvent for the tail region. If `None`, then solvation will be
        performed by the parent `Structure`, using the `Structure.solvent`
        attribute. Other options are coerced to an `SLD` object using
        `SLD(float | complex)`. A float/complex argument is the SLD of the
        solvent (10**-6 Angstrom**-2).
    reverse_monolayer: bool, optional
        The default is to have heads closer to the fronting medium and
        tails closer to the backing medium. If `reverse_monolayer is True`
        then the tails will be closer to the fronting medium and heads
        closer to the backing medium.
    name: str, optional
        The name for the component
    absolute_phi: bool, optional
        Specifies whether `phi_guest_h` and `phi_guest_t` represent fractional
        or absolute volume fractions.
        Please see the notes section for further details.

    Notes
    -----
    The sum of coherent scattering lengths must be in Angstroms, the volume
    must be in cubic Angstroms. This is because the SLD of a tail group is
    calculated as ``b_tails / vm_tails * 1e6`` to achieve the units
    10**6 Angstrom**-2.

    `phi_guest_t` and `phi_guest_h` represent either a fractional (what
    fraction not occupied by lipid is occupied by guest) or absolute
    volume fraction of guest in a layer. If ``absolute_phi is False`` then
    the absolute volume fraction of guest in a layer is calculated as::

        volfrac_guest_x = (1 - volfrac_x) * phi_guest_x

    where ``volfrac_x`` is the absolute volume fraction of lipid in the head
    group region (x representing either the head (h) or tail region (t)).
    The amount of solvent in that layer is then calculated as::

        vfsolv = 1 - volfrac_guest_x - volfrac_x

    If ``absolute_phi is `True`` then `phi_guest_x` is regarded as an
    absolute volume fraction (not relative) and
    ``volfrac_guest_x == phi_guest_x``. ``absolute_phi`` is useful when
    one wants to constrain the volume fraction of guest in the head and
    tail layers to be the same. Otherwise, setting `absolute_phi` to be
    False is preferred, as each layer is less likely to be overfilled.
    Appropriate constraints and bounds must be used during the modelling
    process to ensure ``volfrac_guest_x + volfrac_x <=1``.
    For optimisation using `differential_evolution` this entails
    using the :meth:`LipidLeafletGuest.make_constraints` method to create
    those constraints. Please see `using constraints`_ for an example.
    With MCMC the log-prior term becomes `-np.inf` (i.e. impossible).

    .. _using constraints: https://refnx.readthedocs.io/en/latest/inequality_constraints.html#inequality-constraints-with-differential-evolution
    """

    def __init__(
        self,
        apm,
        b_heads,
        vm_heads,
        thickness_heads,
        b_tails,
        vm_tails,
        thickness_tails,
        rough_head_tail,
        rough_preceding_mono,
        phi_guest_h,
        phi_guest_t,
        sld_guest,
        head_solvent=None,
        tail_solvent=None,
        reverse_monolayer=False,
        name="",
        absolute_phi=False,
    ):
        super().__init__(
            apm,
            b_heads,
            vm_heads,
            thickness_heads,
            b_tails,
            vm_tails,
            thickness_tails,
            rough_head_tail,
            rough_preceding_mono,
            head_solvent=head_solvent,
            tail_solvent=tail_solvent,
            reverse_monolayer=reverse_monolayer,
            name=name,
        )
        warnings.warn(
            "LipidLeafletGuest changed in v0.1.60 to allow guest molecules"
            " in the head region as well. The number AND order of parameters"
            " has changed. Please double check your usage",
            RuntimeWarning,
            stacklevel=2,
        )
        self.phi_guest_h = possibly_create_parameter(phi_guest_h)
        self.phi_guest_h.bounds.lb = 0
        self.phi_guest_h.bounds.ub = 1
        self.phi_guest_t = possibly_create_parameter(phi_guest_t)
        self.phi_guest_t.bounds.lb = 0
        self.phi_guest_t.bounds.ub = 1
        self.sld_guest = possibly_create_scatterer(sld_guest)
        self.absolute_phi = absolute_phi

    def __repr__(self):
        sld_bh = SLD([self.b_heads_real, self.b_heads_imag])
        sld_bt = SLD([self.b_tails_real, self.b_tails_imag])
        s = (
            f"LipidLeaflet("
            f"{self.apm!r}, "
            f"{sld_bh!r}, "
            f"{self.vm_heads!r}, "
            f"{self.thickness_heads!r}, "
            f"{sld_bt!r}, "
            f"{self.vm_tails!r}, "
            f"{self.thickness_tails!r}, "
            f"{self.rough_head_tail!r}, "
            f"{self.rough_preceding_mono!r}, "
            f"{self.phi_guest_h!r}, "
            f"{self.phi_guest_t!r}, "
            f"{self.sld_guest!r}, "
            f"head_solvent={self.head_solvent!r}, "
            f"tail_solvent={self.tail_solvent!r}, "
            f"reverse_monolayer={self.reverse_monolayer}, "
            f"name={self.name!r}, "
            f"absolute_phi={self.absolute_phi!r})"
        )
        return s

    def slabs(self, structure=None):
        """
        Slab representation of monolayer, as an array

        Parameters
        ----------
        structure : refnx.reflect.Structure
            The Structure hosting this Component
        """
        layers = np.zeros((2, 5))
        _sld_guest = complex(self.sld_guest)

        # thicknesses
        layers[0, 0] = float(self.thickness_heads)
        layers[1, 0] = float(self.thickness_tails)

        # sld of guest and tail mixture. water is added on later
        vfh = self.volfrac_h
        vfhg = self.volfrac_guest_h

        re_sld_head = float(self.b_heads_real) / float(self.vm_heads) * 1.0e6
        im_sld_head = float(self.b_heads_imag) / float(self.vm_heads) * 1.0e6

        den = vfh + vfhg
        layers[0, 1] = (vfh * re_sld_head + vfhg * _sld_guest.real) / den
        layers[0, 2] = (vfh * im_sld_head + vfhg * _sld_guest.imag) / den

        # sld of guest and tail mixture. water is added on later
        vft = self.volfrac_t
        vftg = self.volfrac_guest_t

        re_sld_tail = float(self.b_tails_real) / float(self.vm_tails) * 1.0e6
        im_sld_tail = float(self.b_tails_imag) / float(self.vm_tails) * 1.0e6

        den = vft + vftg
        layers[1, 1] = (vft * re_sld_tail + vftg * _sld_guest.real) / den
        layers[1, 2] = (vft * im_sld_tail + vftg * _sld_guest.imag) / den

        # roughnesses
        layers[0, 3] = float(self.rough_preceding_mono)
        layers[1, 3] = float(self.rough_head_tail)

        # volume fractions
        # head region
        # calculate solvation amount
        layers[0, 4] = 1 - vfh - vfhg
        if self.head_solvent is not None:
            _head_solvent = self.head_solvent.complex(
                getattr(structure, "wavelength", None)
            )
            # we do the solvation here, not in Structure.slabs
            layers[0] = overall_sld(layers[0], _head_solvent)
            layers[0, 4] = 0

        # tail region
        # calculate solvation amount
        layers[1, 4] = 1 - vft - vftg
        if self.tail_solvent is not None:
            _tail_solvent = self.tail_solvent.complex(
                getattr(structure, "wavelength", None)
            )
            # we do the solvation here, not in Structure.slabs
            layers[1] = overall_sld(layers[1], _tail_solvent)
            layers[1, 4] = 0

        if self.reverse_monolayer:
            layers = np.flipud(layers)
            layers[:, 3] = layers[::-1, 3]

        return layers

    @property
    def parameters(self):
        p = Parameters(name=self.name)
        p.extend(
            [
                self.apm,
                self.b_heads_real,
                self.b_heads_imag,
                self.vm_heads,
                self.thickness_heads,
                self.b_tails_real,
                self.b_tails_imag,
                self.vm_tails,
                self.thickness_tails,
                self.rough_head_tail,
                self.rough_preceding_mono,
                self.phi_guest_h,
                self.phi_guest_t,
            ]
        )
        p.append(self.sld_guest.parameters)
        if self.head_solvent is not None:
            p.append(self.head_solvent.parameters)
        if self.tail_solvent is not None:
            p.append(self.tail_solvent.parameters)

        return p

    def logp(self):
        # penalise unphysical volume fractions.
        if (
            self.phi_guest_h.value > 1
            or self.phi_guest_t.value > 1
            or self.volfrac_t + self.volfrac_guest_t > 1
            or self.volfrac_h + self.volfrac_guest_h > 1
        ):
            return -np.inf

        return 0

    def make_constraint(self, objective):
        """
        Creates a NonlinearConstraint for a LipidLeafletGuest, ensuring that
        volume fraction of material in the head+tail regions lies in [0, 1].
        Suitable for use by differential_evolution.

        Parameters
        ----------
        objective: refnx.analysis.Objective
            Objective containing the LipidLeafletGuest. Must be the Objective
            that is being minimised by differential_evolution.

        Returns
        -------
        nlc: NonlinearConstraint

        Notes
        -----
        You must create separate constraints for each LipidLeafletGuest object
        in your system.
        The Objective you supply must be for the overall curve fitting system.
        i.e. possibly a GlobalObjective.

        Examples
        --------
        >>> # leaflet is a LipidLeafletGuest, used in an Objective, obj
        >>> con = leaflet.make_constraint(obj)
        >>> fitter = CurveFitter(obj)
        >>> fitter.fit("differential_evolution", constraints=(con,))
        """

        def con(x):
            objective.setp(x)
            return (
                self.volfrac_h + self.volfrac_guest_h,
                self.volfrac_t + self.volfrac_guest_t,
            )

        return NonlinearConstraint(con, 0, 1)

    @property
    def volfrac_guest_h(self):
        """
        Absolute volume fraction of guest in the head group region.
        """
        if self.absolute_phi:
            return self.phi_guest_h.value
        else:
            vfh = self.volfrac_h
            return (1.0 - vfh) * self.phi_guest_h.value

    @property
    def volfrac_guest_t(self):
        """
        Absolute volume fraction of guest in the tail group region.
        """
        if self.absolute_phi:
            return self.phi_guest_t.value
        else:
            vft = self.volfrac_t
            return (1.0 - vft) * self.phi_guest_t.value



from pathlib import Path
import mmap
import os
import warnings
from collections import namedtuple

import numpy as np
import h5py
import pytest

from numpy.testing import assert_equal
import refnx.reduce.event as event
from refnx.reduce import PlatypusNexus, SpatzNexus

try:
    import refnx.reduce._cevent as _cevent
except ImportError:
    HAVE_CEVENTS = False
else:
    HAVE_CEVENTS = True


@pytest.fixture(scope="class")
def event_setup(data_directory):
    if data_directory is None:
        return None

    Event_Setup = namedtuple(
        "Event_Setup",
        [
            "data_directory",
            "event_file_path",
            "nexus_file_path",
            "ReflectNexus",
            "event_list",
            "fpos",
            "f",
            "t",
            "y",
            "x",
        ],
    )

    # PLP example
    event_file_path = Path(
        data_directory,
        "reduce",
        "DAQ_2022-12-02T07-26-10",
        "DATASET_0",
        "EOS.bin",
    )
    nexus_file_path = Path(data_directory, "reduce", "PLP0057863.nx.hdf")
    with open(event_file_path, "rb") as f:
        event_list, fpos = _cevent._cevents(f)

    f, t, y, x = event_list

    stp_PLP = Event_Setup(
        data_directory,
        event_file_path,
        nexus_file_path,
        PlatypusNexus,
        event_list,
        fpos,
        f,
        t,
        y,
        x,
    )

    # SPZ example
    event_file_path = Path(
        data_directory,
        "reduce",
        "DAQ_2022-12-02T14-30-47",
        "DATASET_0",
        "EOS.bin",
    )
    nexus_file_path = Path(data_directory, "reduce", "SPZ0008054.nx.hdf")
    with open(event_file_path, "rb") as f:
        event_list, fpos = _cevent._cevents(f)

    f, t, y, x = event_list

    stp_SPZ = Event_Setup(
        data_directory,
        event_file_path,
        nexus_file_path,
        SpatzNexus,
        event_list,
        fpos,
        f,
        t,
        y,
        x,
    )

    return stp_PLP, stp_SPZ


class TestEvent:
    def test_events_smoke(self, event_setup):
        # check that the event.events function works
        # PLP
        for p in event_setup:
            event.events(p.event_file_path)

    def test_num_events(self, event_setup):
        assert_equal(event_setup[0].x.size, 2209769)
        assert_equal(event_setup[1].x.size, 431693)

    def test_max_frames(self, event_setup):
        # test reading only a certain number of frames

        if HAVE_CEVENTS:
            with open(event_setup[0].event_file_path, "rb") as g:
                event_list, fpos = _cevent._cevents(g, max_frames=1111)
                cyf, cyt, cyy, cyx = event_list

            assert np.max(cyf) < 1111

            event_list, fpos = _cevent._cevents(
                event_setup[1].event_file_path, max_frames=1111
            )
            cyf, cyt, cyy, cyx = event_list
            assert np.max(cyf) < 1111

    def test_event_same_as_detector(self, event_setup):
        # the detector file should be the same as the event file
        # warnings filter for pixel size

        for evt in event_setup:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", RuntimeWarning)

                orig_file = evt.ReflectNexus(evt.nexus_file_path)

            orig_det = orig_file.cat.detector
            frames = [np.arange(0, np.max(evt.f) + 1)]
            event_det, fb = event.process_event_stream(
                evt.event_list,
                frames,
                orig_file.cat.t_bins,
                orig_file.cat.y_bins,
                orig_file.cat.x_bins,
            )
            assert_equal(event_det, orig_det)
            #
            # # PlatypusNexus.process_event_stream should be the same as well
            # det, fc, bm = orig_file.process_event_stream(frame_bins=[])
            # assert_equal(det, orig_det)

    def test_values(self, event_setup):
        # We know the values of all the events in the file from another program
        # test that a set of random events are correct.
        assert_equal(event_setup[0].f[0], 0)
        assert_equal(event_setup[0].t[0], 6904)
        assert_equal(event_setup[0].y[0], 466)
        assert_equal(event_setup[0].x[0], 22)

        assert_equal(event_setup[0].f[-1], 118798)
        assert_equal(event_setup[0].t[-1], 16482)
        assert_equal(event_setup[0].y[-1], 472)
        assert_equal(event_setup[0].x[-1], 16)

    def test_monobloc_events(self, event_setup):
        # the event file changed when the ILL monobloc detector was installed
        data = event.events(event_setup[0].event_file_path)
        f, t, y, x = data[0]
        assert len(t) == 2209769

        # in the nexus setup dcr used
        x_bins = np.array([9.5, 19.5])
        y_bins = np.linspace(-0.5, 1023.5, 1025)
        t_bins = np.linspace(0, 30000, 1001)
        frames = [np.arange(0, 118800, 1)]

        detector, fbins = event.process_event_stream(
            data[0], frames, t_bins, y_bins, x_bins
        )

        with h5py.File(event_setup[0].nexus_file_path, "r") as g:
            orig_det = g["entry1/data/hmm"][:]
            assert orig_det.shape == (1, 1000, 1024, 1)

        assert_equal(detector, orig_det)

    def test_clock_scale(self, event_setup):
        # the event file changed when the ILL monobloc detector was installed
        pth = Path(event_setup[0].event_file_path)

        # PLP clock_scale is 16 ns per tick
        with open(pth, "rb") as f:
            buffer = mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ)
        hdr_base, hdr_packing = _cevent.event_header(buffer)
        assert hdr_base.clock_scale == 16

        # SPZ clock_scale is 100 ns per tick
        pth = Path(event_setup[1].event_file_path)
        with open(pth, "rb") as f:
            buffer = mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ)
        hdr_base, hdr_packing = _cevent.event_header(buffer)
        assert hdr_base.clock_scale == 100



#! /usr/bin/env python
# System imports
from setuptools import setup, Extension
import os
import subprocess
import platform
import sys
import warnings
import glob
import tempfile
import textwrap
import subprocess

try:
    from Cython.Build import cythonize
except ImportError:
    USE_CYTHON = False
    warnings.warn(
        "Cython was not found. Slow reflectivity calculations will be used."
    )
else:
    USE_CYTHON = True


# versioning
MAJOR = 0
MINOR = 1
MICRO = 61
ISRELEASED = True
VERSION = f"{MAJOR}.{MINOR}.{MICRO}"


# are we on windows, darwin, etc?
platform = sys.platform
# packages = find_packages()
# try:
#     idx = packages.index("benchmarks")
#     if idx >= 0:
#         packages.pop(idx)
#     idx = packages.index("benchmarks.benchmarks")
#     if idx >= 0:
#         packages.pop(idx)
#     idx = packages.index("motofit")
#     if idx >= 0:
#         packages.pop(idx)
# except ValueError:
#     pass


# Return the git revision as a string
def git_version():
    def _minimal_ext_cmd(cmd):
        # construct minimal environment
        env = {}
        for k in ["SYSTEMROOT", "PATH"]:
            v = os.environ.get(k)
            if v is not None:
                env[k] = v
        # LANGUAGE is used on win32
        env["LANGUAGE"] = "C"
        env["LANG"] = "C"
        env["LC_ALL"] = "C"
        out = subprocess.Popen(
            cmd, stdout=subprocess.PIPE, env=env
        ).communicate()[0]
        return out

    try:
        out = _minimal_ext_cmd(["git", "rev-parse", "HEAD"])
        GIT_REVISION = out.strip().decode("ascii")
    except OSError:
        GIT_REVISION = "Unknown"

    return GIT_REVISION


def get_version_info():
    # Adding the git rev number needs to be done inside
    # write_version_py(), otherwise the import of refnx.version messes
    # up the build under Python 3.
    FULLVERSION = VERSION
    if os.path.exists(".git"):
        GIT_REVISION = git_version()
    elif os.path.exists("refnx/version.py"):
        # must be a source distribution, use existing version file
        # load it as a separate module to not load refnx/__init__.py
        import importlib.util

        spec = importlib.util.spec_from_file_location(
            "version", "refnx/version.py"
        )
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        GIT_REVISION = module.git_revision
    else:
        GIT_REVISION = "Unknown"

    if not ISRELEASED:
        FULLVERSION += ".dev0+" + GIT_REVISION[:7]

    return FULLVERSION, GIT_REVISION


def write_version_py(filename="refnx/version.py"):
    cnt = """
# THIS FILE IS GENERATED FROM REFNX SETUP.PY
short_version = '%(version)s'
version = '%(version)s'
full_version = '%(full_version)s'
git_revision = '%(git_revision)s'
release = %(isrelease)s
if not release:
    version = full_version
"""
    FULLVERSION, GIT_REVISION = get_version_info()

    a = open(filename, "w")
    try:
        a.write(
            cnt
            % {
                "version": VERSION,
                "full_version": FULLVERSION,
                "git_revision": GIT_REVISION,
                "isrelease": str(ISRELEASED),
            }
        )
    finally:
        a.close()


###############################################################################
"""
Is openMP usable?
"""
CCODE = textwrap.dedent(
    """\
    #include <omp.h>
    #include <stdio.h>
    int main(void) {
    #pragma omp parallel
    printf("nthreads=%d\\n", omp_get_num_threads());
    return 0;
    }
    """
)


def get_openmp_flag(compiler):
    if hasattr(compiler, "compiler"):
        compiler = compiler.compiler[0]
    else:
        compiler = compiler.__class__.__name__

    if sys.platform == "win32" and ("icc" in compiler or "icl" in compiler):
        return ["/Qopenmp"], []
    elif sys.platform == "win32":
        return ["/openmp"], []
    elif sys.platform == "darwin" and ("icc" in compiler or "icl" in compiler):
        return ["-openmp"], []
    elif sys.platform == "darwin":
        # default for macOS, assuming Apple-clang
        # -fopenmp can't be passed as compile flag when using Apple-clang.
        # OpenMP support has to be enabled during preprocessing.
        #
        # it may be possible that someone builds with a different/updated
        # compiler (don't know how to check for that).
        #
        # set the following environment variables, assumes that llvm openmp
        # has been built and installed by the user.
        #
        # brew install libomp
        # export CC=clang
        # export CXX =clang++
        # export CXXFLAGS="$CXXFLAGS -Xpreprocessor -fopenmp"
        # export CFLAGS="$CFLAGS -I/usr/local/opt/libomp/include"
        # export CXXFLAGS="$CXXFLAGS -I/usr/local/opt/libomp/include"
        # export LDFLAGS="$LDFLAGS -L/usr/local/opt/libomp/lib -lomp"
        # export DYLD_LIBRARY_PATH =/usr/local/opt/libomp/lib
        return ["-fopenmp"], ["-lomp"]
    # Default flag for GCC and clang:
    return ["-fopenmp"], []


def check_openmp_support():
    """Check whether OpenMP test code can be compiled and run"""

    try:
        from setuptools._distutils.ccompiler import new_compiler
        from setuptools._distutils.sysconfig import customize_compiler

        # from numpy.distutils.ccompiler import new_compiler
        # from distutils.sysconfig import customize_compiler
        from distutils.errors import CompileError, LinkError
    except ImportError:
        return False

    ccompiler = new_compiler()
    customize_compiler(ccompiler)

    start_dir = os.path.abspath(".")

    with tempfile.TemporaryDirectory() as tmp_dir:
        try:
            os.chdir(tmp_dir)

            # Write test program
            with open("test_openmp.c", "w") as f:
                f.write(CCODE)

            os.mkdir("objects")

            # Compile, test program
            openmp_flags = get_openmp_flag(ccompiler)

            ccompiler.compile(
                ["test_openmp.c"],
                output_dir="objects",
                extra_postargs=openmp_flags[0],
            )

            # Link test program
            extra_preargs = os.getenv("LDFLAGS", None)
            if extra_preargs is not None:
                extra_preargs = extra_preargs.split(" ")
            else:
                extra_preargs = []

            objects = glob.glob(
                os.path.join("objects", "*" + ccompiler.obj_extension)
            )
            ccompiler.link_executable(
                objects,
                "test_openmp",
                extra_preargs=extra_preargs,
                extra_postargs=openmp_flags[1],
            )

            # Run test program
            output = subprocess.check_output("./test_openmp")
            output = output.decode(sys.stdout.encoding or "utf-8").splitlines()

            # Check test program output
            if "nthreads=" in output[0]:
                nthreads = int(output[0].strip().split("=")[1])
                openmp_supported = len(output) == nthreads
            else:
                openmp_supported = False

        except (CompileError, LinkError, subprocess.CalledProcessError):
            openmp_supported = False

        finally:
            os.chdir(start_dir)

    return openmp_supported


# do you want to parallelise things with openmp?
HAS_OPENMP = check_openmp_support()
# HAS_OPENMP = False
###############################################################################


# refnx setup
info = {
    "include_package_data": True,
}


####################################################################
# this is where setup starts
####################################################################
def setup_package():

    # Rewrite the version file every time
    write_version_py()
    info["version"] = get_version_info()[0]
    print(info["version"])

    if USE_CYTHON:
        # Obtain the numpy include directory.  This logic works across numpy
        # versions.
        ext_modules = []
        HAS_NUMPY = True

        try:
            import numpy as np
        except:
            info["setup_requires"] = ["numpy"]
            HAS_NUMPY = False

        if HAS_NUMPY:
            try:
                numpy_include = np.get_include()
            except AttributeError:
                numpy_include = np.get_numpy_include()

            _cevent = Extension(
                name="refnx.reduce._cevent",
                sources=["src/_cevent.pyx"],
                include_dirs=[numpy_include],
                language="c++",
                define_macros=[
                    ("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")
                ],
                # libraries=
                # extra_compile_args = "...".split(),
            )
            ext_modules.append(_cevent)

            _cutil = Extension(
                name="refnx._lib._cutil",
                sources=["src/_cutil.pyx"],
                include_dirs=[numpy_include],
                language="c",
                # libraries=
                # extra_compile_args = "...".split(),
            )
            ext_modules.append(_cutil)

            # creflect extension module
            # Compile reflectivity calculator to object with C compiler
            # first.
            # It's not possible to do this in an Extension object because
            # the `-std=c++11` compile arg and C99 C code are incompatible
            # (at least on Darwin).
            from setuptools._distutils.ccompiler import new_compiler
            from setuptools._distutils.sysconfig import customize_compiler

            # from numpy.distutils.ccompiler import new_compiler
            # from distutils.sysconfig import customize_compiler

            ccompiler = new_compiler()
            customize_compiler(ccompiler)
            ccompiler.verbose = True
            extra_preargs = [
                "-O3",
            ]

            if sys.platform == "win32":
                # use the C++ code on Windows. The C++ code uses the
                # std::complex<double> object for its arithmetic.
                f = ["src/refcalc.cpp"]
            else:
                # and C code on other machines. The C code uses C99 complex
                # arithmetic which is 10-20% faster.
                # the CMPLX macro was only standardised in C11
                extra_preargs.extend(
                    [
                        "-std=c17",
                    ]
                )
                f = ["src/refcalc.c"]
            refcalc_obj = ccompiler.compile(f, extra_preargs=extra_preargs)
            # print(refcalc_obj)

            _creflect = Extension(
                name="refnx.reflect._creflect",
                sources=[
                    "src/_creflect.pyx",
                    "src/refcaller.cpp",
                    "src/pnr/magnetic.cc",
                ],
                include_dirs=[numpy_include],
                language="c++",
                extra_compile_args=["-std=c++11"],
                define_macros=[
                    ("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")
                ],
                extra_objects=refcalc_obj,
            )
            ext_modules.append(_creflect)

            # if we have openmp use pure cython version
            # openmp should be present on windows, linux
            #
            # However, it's not present in Apple Clang. Therefore one has to
            # jump through hoops to enable it.
            # It's probably easier to install OpenMP on macOS via homebrew
            #
            # brew install llvm
            # export PATH="/opt/homebrew/opt/llvm/bin:$PATH"
            # export CPPFLAGS="-I/opt/homebrew/opt/llvm/include"
            # export LDFLAGS="-L/opt/homebrew/opt/llvm/lib/c++ -Wl,-rpath,/opt/homebrew/opt/llvm/lib/c++ -L/opt/homebrew/opt/llvm/lib"
            print(f"{HAS_OPENMP=}, {refcalc_obj=}")
            if HAS_OPENMP:
                # cyreflect extension module
                _cyreflect = Extension(
                    name="refnx.reflect._cyreflect",
                    sources=[
                        "src/_cyreflect.pyx",
                        "src/refcaller.cpp",
                        "src/pnr/magnetic.cc",
                    ],
                    include_dirs=[numpy_include],
                    language="c++",
                    extra_compile_args=[],
                    extra_link_args=[],
                    define_macros=[],
                    extra_objects=refcalc_obj,
                    # extra_compile_args = "...".split(),
                )
                openmp_flags = get_openmp_flag(ccompiler)
                _cyreflect.extra_compile_args += openmp_flags[0]
                _cyreflect.extra_link_args += openmp_flags[1]

                ext_modules.append(_cyreflect)

            # specify min deployment version for macOS
            # if platform == "darwin":
            #     for mod in ext_modules:
            #         mod.extra_compile_args.append("-mmacosx-version-min=10.9")

            info["ext_modules"] = cythonize(ext_modules)
            info["zip_safe"] = False

    try:
        setup(**info)
    except ValueError:
        # there probably wasn't a C-compiler (windows). Try removing extension
        # compilation
        print("")
        print("*****WARNING*****")
        print(
            "You didn't try to build the Reflectivity calculation extension."
            " Calculation will be slow, falling back to pure python."
            " To compile extension install cython. If installing in windows you"
            " should then install from Visual Studio command prompt (this makes"
            " C compiler available"
        )
        print("*****************")
        print("")
        info.pop("cmdclass")
        info.pop("ext_modules")
        setup(**info)


if __name__ == "__main__":
    setup_package()



import sys
import logging
import time
from pathlib import Path


def gui():
    from qtpy import QtWidgets
    from refnx.reduce._app.view import SlimWindow

    time_str = time.strftime("%Y%m%d-%H%M%S")
    log_filename = "slim_" + time_str + ".log"
    log_filename = Path.home() / log_filename

    logging.basicConfig(
        filename=log_filename,
        level=logging.INFO,
        format="%(asctime)s %(message)s",
    )
    logging.info("Starting SLIM reduction")

    app = QtWidgets.QApplication(sys.argv)
    myapp = SlimWindow()

    myapp.show()
    v = app.exec()
    return v


def main():
    sys.exit(gui())


__all__ = [gui, main]



import numpy as np
from numpy.testing import assert_allclose

from refnx.analysis import Parameter
from refnx.reflect import FunctionalForm, SLD


class Line:
    def __call__(self, z, extent, left_sld, right_sld, **kwds):
        self.keys = list(kwds.keys())
        self.extent = extent

        grad = (right_sld - left_sld) / extent
        intercept = left_sld
        # we don't calculate the volume fraction of solvent
        return z * grad * kwds["dummy_param"] + intercept, None


def quadratic(z, extent, left_sld, right_sld, x=None, y=None):
    res = np.polyfit(
        [0.0, x, extent], [np.real(left_sld), y, np.real(right_sld)], deg=2
    )
    return np.polyval(res, z), None


def test_functional_form():
    si = SLD(2.07)
    d2o = SLD(6.36)
    p = Parameter(1.0)
    line = Line()

    form = FunctionalForm(100, line, dummy_param=p)
    s = si | form | d2o(0, 3)
    s.sld_profile()

    assert_allclose(line.extent, 100.0)
    assert "dummy_param" in line.keys

    x = Parameter(4.0)
    y = Parameter(5.0)
    quad = FunctionalForm(100.0, quadratic, x=x, y=y)
    s = si | quad | d2o(0, 3)
    s.slabs()
